#include "Console.hpp"
#include "algebra.hpp"
ConsoleItem::ConsoleItem()
{
	m_strline = "";
	m_col = Colour(0,0,0);
	m_type = InfoType;	
}


ConsoleItem::ConsoleItem(std::string str, ItemType itype)
{
	m_strline = str;
	m_type = itype;
	if (m_type == CommandType)
		m_col = Colour(1,1,1);
	else if (m_type == InfoType)
		m_col = Colour(0,1,1);
	else if (m_type == ErrorType)
		m_col = Colour(1,0,0);
	else if (m_type == CommentType)
		m_col = Colour(0,0.8,0);
	m_linenum = 0;
}

ConsoleItem::ConsoleItem(std::string str, int num, ItemType itype)
{
	m_strline = str;
	m_type = itype;
	if (m_type == CommandType)
		m_col = Colour(1,1,1);
	else if (m_type == InfoType)
		m_col = Colour(0,1,1);
	else if (m_type == ErrorType)
		m_col = Colour(1,0,0);
	else if (m_type == CommentType)
		m_col = Colour(0,0.8,0);
	m_linenum = num;
}

ConsoleItem::~ConsoleItem() { }


std::string ConsoleItem::getString()
{
	return m_strline;
}

Colour ConsoleItem::getColour()
{
	return m_col;
}

ItemType ConsoleItem::getType()
{
	return m_type;
}













Console::Console()
{
	consoleline = 0;
	offsetcon = 0;
	currentinput = "";
	lastcmdindex = 0;
	echoon = true;
	consoleState = ConsoleUp;
	consoleDisplayOffset = -150;
	consoleDelay = 5;
	lastCommand = "";
    displayLast = false;
    displayLastTimer = 0;


}


Console::~Console() 
{
}


inline std::string Console::convertToString(int i)
{
   std::ostringstream o;
   o << i;
   std::string x = o.str();
   return x;
   
} 

inline std::string Console::convertToString(double d)
{
   std::ostringstream o;
   o << d;
   std::string x = o.str();
   return x;
   
} 


std::string Console::trim(std::string s)
{
	
	while(true)
	{
		if (s.length() == 0) break;
		if (isspace(s.at(0)))
		{
			s.erase(s.begin());
		}
		else
		{
			break;
		}
	}
	if (s.length() > 0)
	{
		while(true)
		{
			if (s.length() == 0) break;
			if (isspace(s.at(s.length()-1))) 
			{
				s.erase(s.end()-1);
			}
			else
			{
				break;
			}
		}
	}
	return s;
}


void Console::drawConsole()
{

        if (displayLast)
        {
           glColor3d(displayItem.getColour().R(), displayItem.getColour().G(), displayItem.getColour().B());
           glPrint((640-10*displayItem.getString().length())/2, 30, const_cast<char *>(displayItem.getString().c_str()), 1, 0 );
           if (SDL_GetTicks() > displayLastTimer)
           {
              displayLast = false;
              displayLastTimer = 0;
           }
        }
		


        if (consoleState == ConsoleUp) return;


		if (consoleState == ConsoleGoingDown)
		{
			if (SDL_GetTicks() >= consoleDelay)
			{
				consoleDisplayOffset+=10;

				if (consoleDisplayOffset == 0)
				{
					consoleState = ConsoleDown;
          		}
			}

		}
		else if (consoleState == ConsoleGoingUp)
		{
			if (SDL_GetTicks() >= consoleDelay)
			{
			
				consoleDisplayOffset-=10;

				if (consoleDisplayOffset == -(200))
				{
					consoleState = ConsoleUp;
				}
			}


		}

		

    glBindTexture( GL_TEXTURE_2D, backtext );

    glDisable( GL_DEPTH_TEST );
    glDisable( GL_LIGHTING );
    glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA,GL_SRC_ALPHA);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    //glBlendFunc(GL_SRC_ALPHA,GL_ONE);

    glMatrixMode( GL_PROJECTION );
    glPushMatrix( );

    glLoadIdentity( );
    glOrtho( 0, 640, 0, 480, -1, 1 );

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix( );
    glLoadIdentity( );

    glTranslated(0, -consoleDisplayOffset, 0); 
    glColor3d(1,1,1);
    glBegin( GL_QUADS );
   
    glTexCoord2f( 0, (double)150/256 );
    glVertex2i( 0, 480-150 );

    glTexCoord2f( (double)640/1024, (double)150/256 );
    glVertex2i( 640, 480-150 );

    glTexCoord2f( (double)640/1024, 0 );    
    glVertex2i( 640, 480 );
    
    
    glTexCoord2f( 0, 0 );
    glVertex2i( 0, 480 );
    
  
    glEnd();

    glMatrixMode( GL_PROJECTION );
    glPopMatrix( );

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix( );

    glDisable(GL_BLEND);
    glEnable( GL_LIGHTING );
    glEnable( GL_DEPTH_TEST );








		int numlines = NUMCONSOLELINE;
		for (int i = 0; i < numlines; i++)
			{
				if (offsetcon + i < console.size()) 
				{
					std::string strlinenumber = "";

					int intlinenumber = console[offsetcon + i].getLineNumber();
					if (intlinenumber < 10)
						strlinenumber = strlinenumber + convertToString(intlinenumber) + "  :";
					else if (intlinenumber < 100)
						strlinenumber = strlinenumber + convertToString(intlinenumber) + " :";
					else
						strlinenumber = strlinenumber + convertToString(intlinenumber) + ":";
					
					//if (console[offsetcon + i].getType() == CommandType)
					//	surf->TextXY(5, 5+13*i + consoleDisplayOffset, console[offsetcon + i].getColour(), std::string(strlinenumber + "-> " + console[offsetcon + i].getString()).c_str());
					//else
					//	surf->TextXY(5, 5+13*i + consoleDisplayOffset, console[offsetcon + i].getColour(), std::string(strlinenumber + console[offsetcon + i].getString()).c_str());
                    glColor3d(console[offsetcon + i].getColour().R(), console[offsetcon + i].getColour().G(), console[offsetcon + i].getColour().B());
                    if (console[offsetcon + i].getType() == CommandType)
		      glPrint( 5, 460-(5+13*i+ consoleDisplayOffset), const_cast<char *>(std::string(strlinenumber + std::string("-> ") + console[offsetcon + i].getString()).c_str()), 1, 0 );
		    else
		      glPrint( 5, 460-(5+13*i+ consoleDisplayOffset), const_cast<char *>(std::string(strlinenumber + console[offsetcon + i].getString()).c_str()), 1, 0 );
		      
				}
			}
			
			std::string t = "> " + currentinput + "|";
            glColor3d(1,1,1);
            glPrint( 5, 460-(5+14*9+ consoleDisplayOffset)+5, const_cast<char *>(t.c_str()), 1, 0 );
			
			

	
}


void Console::update(SDL_KeyboardEvent key)
{
	
	int numlines = NUMCONSOLELINE;
    
    if (key.state == SDL_PRESSED)
    {
       if (key.keysym.sym == SDLK_RETURN)
       {
		  currentinput = trim(currentinput);
          if (currentinput.length() > 0)
		  {
		    offsetcon = console.size() - numlines;
			lastCommand = currentinput;
			currentinput = "";
			lastcmdindex = 0;
		  }

	   }
       else if (key.keysym.sym == SDLK_UP)
       {
		 offsetcon--;
		 if (offsetcon < 0) offsetcon = 0;
       }
       else if (key.keysym.sym == SDLK_DOWN)
       {
		 if (numlines <= console.size()) offsetcon++;
		 if (offsetcon > console.size()-numlines) offsetcon = console.size() - numlines;
       }	
	
	
       //fix all this so it cycles properly
       else if (key.keysym.sym == SDLK_LEFT)
       {
		
		std::string oldinput = currentinput;
		int oldlastcmdindex = lastcmdindex;

		std::vector<ConsoleItem>::reverse_iterator it = console.rbegin();
		it = it + lastcmdindex;

		bool found = false;

		for (; it != console.rend(); it++)
		{
			lastcmdindex++;
			if ( it->getType() == CommandType )
			{
				currentinput = it->getString();
				found = true;
				break;
			}
		}

		
		if (!found)
		{
			currentinput = oldinput;
			lastcmdindex = oldlastcmdindex;
		}
      }
      else if (key.keysym.sym == SDLK_RIGHT)
      {
		
		bool found = false;
		std::vector<ConsoleItem>::iterator it = console.begin();
		
		if (lastcmdindex == 0)
			it = console.end();
		else
			it = it + (console.size() - lastcmdindex) +1;
		
		for (; it != console.end(); it++)
		{
			lastcmdindex--;
			if ( it->getType() == CommandType )
			{
				currentinput = it->getString();
				found = true;
				break;
			}
		}

		if (!found)
		{
			lastcmdindex = 0;
			currentinput = "";
		}

     }
    }

	char messageChar = key.keysym.unicode & 0x7F;
	if (messageChar != 0 && key.state == SDL_PRESSED)
	{
		
		
		if (messageChar == 0x08) //backspace
		{
			currentinput = currentinput.substr(0, currentinput.size()-1);
		}

		else
		{
			std::string alphanum = " 1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!@#$%^&*()-_=+,<.>/?;:'[{]}|";
			if (alphanum.find(messageChar) != std::string::npos)
			{
				currentinput = currentinput + messageChar;
				
			}
		}	
		messageChar = 0;
	}

}



void Console::toggle()
{
				
	if (consoleState == ConsoleUp)
	{
        consoleDelay = SDL_GetTicks() + 5;
		consoleState = ConsoleGoingDown;
	}
	else if (consoleState == ConsoleGoingDown)
	{
		consoleState = ConsoleGoingUp;
	}
	else if (consoleState == ConsoleGoingUp)
	{
		consoleState = ConsoleGoingDown;
	}
	else if (consoleState == ConsoleDown)
	{
        consoleDelay = SDL_GetTicks() + 5;
		consoleState = ConsoleGoingUp;
	}
}



void Console::writeConsole(ConsoleItem conitem)
{
	consoleline++;

	if (console.size() == 500)
	{
		console.erase(console.begin());
	}

	ConsoleItem ci = ConsoleItem(conitem.getString(), consoleline, conitem.getType());
	console.push_back(ci);

	//consolecol.push_back(col);
	setConsoleToEnd();

}

void Console::writeConsoleDisp(ConsoleItem conitem)
{
	consoleline++;

	if (console.size() == 500)
	{
		console.erase(console.begin());
	}

	ConsoleItem ci = ConsoleItem(std::string("MSG: ") + conitem.getString(), consoleline, conitem.getType());
	console.push_back(ci);

	//consolecol.push_back(col);
	setConsoleToEnd();

    displayLast = true;
    displayLastTimer = SDL_GetTicks() + 2000;
    displayItem = conitem;
    
}



void Console::setConsoleToEnd()
{
	int numlines = NUMCONSOLELINE;
	offsetcon = console.size() - numlines;
}

bool Console::echoOn()
{
	return echoon;
}

void Console::setEcho(bool ech)
{
	echoon = ech;
}

std::string Console::getLastCommand()
{
	std::string temp = lastCommand;
	lastCommand = "";
	return temp;
}

GLvoid Console::glPrint( GLint x, GLint y, char *string, double scalef, int set )
{
    if ( set > 1 )
	set = 1;

    /* Select our texture */
    glBindTexture( GL_TEXTURE_2D, fonttext );

    /* Disable depth testing */
    glDisable( GL_DEPTH_TEST );
    glDisable( GL_LIGHTING );
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    /* Select The Projection Matrix */
    glMatrixMode( GL_PROJECTION );
    /* Store The Projection Matrix */
    glPushMatrix( );

    /* Reset The Projection Matrix */
    glLoadIdentity( );
    /* Set Up An Ortho Screen */
    glOrtho( 0, 640, 0, 480, -1, 1 );

    /* Select The Modelview Matrix */
    glMatrixMode( GL_MODELVIEW );
    /* Stor the Modelview Matrix */
    glPushMatrix( );
    /* Reset The Modelview Matrix */
    glLoadIdentity( );

    /* Position The Text (0,0 - Bottom Left) */
    glTranslated( x, y, 0 );
    
    glScaled(scalef, scalef, scalef);
    
    /* Choose The Font Set (0 or 1) */
    glListBase( fontbase - 32 + ( 128 * set ) );

    /* Write The Text To The Screen */
    glCallLists( strlen( string ), GL_BYTE, string );
    
    /* Select The Projection Matrix */
    glMatrixMode( GL_PROJECTION );
    /* Restore The Old Projection Matrix */
    glPopMatrix( );

    /* Select the Modelview Matrix */
    glMatrixMode( GL_MODELVIEW );
    /* Restore the Old Projection Matrix */
    glPopMatrix( );

    /* Re-enable Depth Testing */
    glDisable(GL_BLEND);
    glEnable( GL_LIGHTING );
    glEnable( GL_DEPTH_TEST );
}



