#ifndef CONSOLE_H_XIX
#define CONSOLE_H_XIX


#include <iostream>
#include <string>
#include "algebra.hpp"
#include <SDL/SDL.h>
#include <sstream>
#include <vector>
#include <GL/gl.h>
#include <GL/glu.h>


enum ItemType
{
	CommandType,
	InfoType,
	ErrorType,
	CommentType
};

class ConsoleItem
{
public:
	ConsoleItem();
	ConsoleItem(std::string str, ItemType itype);
	ConsoleItem(std::string str, int num, ItemType itype);
	virtual ~ConsoleItem();
	std::string getString();
	Colour getColour();
	ItemType getType();
	int getLineNumber() { return m_linenum; }
    
private:
	std::string m_strline;
	Colour m_col;
	ItemType m_type;
	int m_linenum;
	
};


#define NUMCONSOLELINE 9;


enum ConsoleStateEnum
{
	ConsoleGoingDown,
	ConsoleGoingUp,
	ConsoleDown,
	ConsoleUp
};

class Console
{
public:
	Console();
	virtual ~Console();
	void drawConsole();
	void update(SDL_KeyboardEvent key);
	void writeConsole(ConsoleItem conitem);
    void writeConsoleDisp(ConsoleItem conitem);

	void setConsoleToEnd();
	void toggle();
	
	bool echoOn();
	void setEcho(bool ech);
	std::string getLastCommand();

    void setFontTexture(GLuint f, GLuint backt, GLuint b) { fonttext = f; backtext = backt; fontbase = b;}
    ConsoleStateEnum state() { return consoleState; }

private:
	std::vector<ConsoleItem> console;
	int consoleline;
	int offsetcon;
	std::string currentinput;
	int lastcmdindex;
	bool echoon;
	ConsoleStateEnum consoleState;
	int consoleDisplayOffset;
	int consoleDelay;

	inline std::string convertToString(int i);
	inline std::string convertToString(double d);
	std::string trim(std::string s);
    GLvoid glPrint( GLint x, GLint y, char *string, double scalef, int set );

	std::string lastCommand;

    bool displayLast;
    Uint32 displayLastTimer;
    ConsoleItem displayItem;
    GLuint fonttext;
    GLuint backtext;

    GLuint fontbase;


	
};

#endif

