#include "a2.hpp"

// Return a matrix to represent a counterclockwise rotation of "angle"
// degrees around the axis "axis", where "axis" is one of the
// characters 'x', 'y', or 'z'.
Matrix4x4 rotation(double angle, char axis)
{
  Matrix4x4 r;  //this is identity
  if (axis == 'x')
  { 
    r[1][1] = cos(angle);
    r[1][2] = -sin(angle);
    r[2][1] = sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'y')
  { 
    r[0][0] = cos(angle);
    r[0][2] = sin(angle);
    r[2][0] = -sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'z')
  { 
    r[0][0] = cos(angle);
    r[0][1] = -sin(angle);
    r[1][0] = sin(angle);
    r[1][1] = cos(angle);

  }

  //if invalid axis, just uses identity


  return r;
}

// Return a matrix to represent a displacement of the given vector.
Matrix4x4 translation(const Vector3D& displacement)
{
  Matrix4x4 t;
  for (int i = 0; i < 3; i++)
    t[i][3] = displacement[i];
  return t;
}

// Return a matrix to represent a nonuniform scale with the given factors.
Matrix4x4 scaling(const Vector3D& scale)
{
  Matrix4x4 s;
  for (int i = 0; i < 3; i++)
    s[i][i] = scale[i];
  return s;
}

double degToRad(double ang) { return (M_PI/(double)180) * ang; }
double radToDeg(double ang) { return ((double)180/M_PI) * ang; }

