#include "entity.hpp"
#include "scene.hpp"
#include "viewer.hpp"
    
Enemy::Enemy(std::string name, SceneNode *node, std::map<int, std::map<std::string, animElement> > animMap, int frameMax, Point3D pos, int life)
{
 this->enemyPos = pos;
 this->name = name;
 this->animMap = animMap;  
 this->enemyPos = pos;
 this->life = life;
 this->frameMax = frameMax;
 this->obj = node;
 this->maxlife = life;
 frameNum = 0;
 dead = false;
 lastShotTimer = 0;
}

Enemy::Enemy() 
{ }

Enemy::~Enemy()
{ }
    
void Enemy::updateOrientation(Vector3D dir, Vector3D up)
{ 
  Vector3D z = dir; 
  z.normalize();
  Vector3D x = up.cross(z);
  x.normalize(); 
  Vector3D y = z.cross(x);
  Matrix4x4 m;

  m[0][0] = x[0];
  m[0][1] = x[1];
  m[0][2] = x[2];
  m[1][0] = y[0];
  m[1][1] = y[1];
  m[1][2] = y[2];
  m[2][0] = z[0];
  m[2][1] = z[1];
  m[2][2] = z[2];

  enemyRotation = m.transpose();
} 
    
EnemyActivity Enemy::tick(Point3D mypos)
{
  if (!dead)
  {    
      animTick();
      double lengthToMe = (mypos-enemyPos).length();
      if (lengthToMe < 350 && (lengthToMe > 50 || ((double)life/maxlife) < 0.4) )
      {
        if (rand()%3==0) enemyVelocity[2] = enemyVelocity[2] + 0.02 + 0.01*enemyVelocity[2];
        if (enemyVelocity[2] > 2) enemyVelocity[2] = 2;
      }
      decelerate(enemyVelocity);
      enemyPos = enemyPos + enemyRotation*enemyVelocity;
      if (life <= 0) 
      { 
         dead = true; 
         life = 0;
         if (rand()%2 == 0) return ENEMY_DROP_PRESENT;  //return a present
         return ENEMY_JUST_DIED;
      }
      if (lengthToMe < 100 && SDL_GetTicks() > lastShotTimer)
      {
        lastShotTimer = SDL_GetTicks() + 2000;
	if (rand()%2 == 0)
	   return ENEMY_SHOOT; //shoot at me      
      }
      
      return ENEMY_ALIVE; //all is well
  }
  else
  { 
      return ENEMY_DEAD; //dead, do nothing
  
  }
}

void Enemy::render()
{
    glColor3d(1,1,1);
    glPushMatrix();
      glTranslated(enemyPos[0],enemyPos[1],enemyPos[2]);
      glPushMatrix();
        glMultMatrixd(enemyRotation.transpose().begin());
        glRotated(90+180, 1,0,0);	
        glScaled(0.4, 0.4, 0.4); 
        obj->walk_gl(0);
      glPopMatrix();
    glPopMatrix();
}

void Enemy::doDamage(int amt)
{
  life = life - amt;
  if (life < 0) life = 0;
}

void Enemy::decelerate(Vector3D &v)
{
    if (v[2] > 0)
      v[2] = v[2] - 0.02 * v[2];
    if (v[2] < 0)
      v[2] = v[2] - 0.02 * v[2];
    if (v[2] > -0.01 && v[2] < 0.01) v[2] = 0;

    if (v[0] > 0)
      v[0] = v[0] - 0.02 * v[0];
    if (v[0] < 0)
      v[0] = v[0] - 0.02 * v[0];
    if (v[0] > -0.01 && v[0] < 0.01) v[0] = 0;
    

    if (v[1] > 0)
      v[1] = v[1] - 0.02 * v[1];
    if (v[1] < 0)
      v[1] = v[1] - 0.02 * v[1];
    if (v[1] > -0.01 && v[1] < 0.01) v[1] = 0;
    

}


void Enemy::animTick()
{
  if (frameNum == frameMax)
  {
    frameNum = 0;
  }
  frameNum++;
  for (std::map<int, SceneNode*>::iterator it = SceneNode::nodeMap[name].begin(); it != SceneNode::nodeMap[name].end(); it++)
  {
    JointNode *jointNode = dynamic_cast<JointNode*>((*it).second);
    if (jointNode != 0)
    {
      std::string jointName = jointNode->get_name();  
      std::map<std::string, animElement> keyframe = animMap[frameNum];
      animElement *ae = &keyframe[jointName];
      jointNode->set_joint_x(ae->anglex);
      jointNode->set_joint_y(ae->angley);
    }
  }
}









