#ifndef ENTITY_XIX_H
#define ENTITY_XIX_H

#include "viewer.hpp"
#include "algebra.hpp"
#include <SDL/SDL.h>

struct animElement;

enum EnemyActivity
{
  ENEMY_ALIVE,
  ENEMY_SHOOT,
  ENEMY_DEAD,
  ENEMY_DROP_PRESENT,
  ENEMY_JUST_DIED
};

class Enemy
{

  public:
    Enemy(std::string name, SceneNode *node, std::map<int, std::map<std::string, animElement> > animMap, int frameMax, Point3D pos, int life);
    Enemy();
    ~Enemy();
    
    void updateOrientation(Vector3D dir, Vector3D up);
    EnemyActivity tick(Point3D mypos);
    Point3D getPos() { return enemyPos; }
    Vector3D getVel() { return enemyVelocity; }
    void setPos(Point3D p) { enemyPos = p; }
    void setVel(Vector3D v) { enemyVelocity = v; }
    Vector3D getFullVel() { return enemyRotation * enemyVelocity; }
    void render();
    void doDamage(int amt);

    bool alive() {return !dead;}
    void kill() {life = 0;}
    
  private:

    void decelerate(Vector3D &v);
    void animTick();
  
    SceneNode* obj;
  
    Point3D enemyPos;
    Vector3D enemyVelocity;
    Matrix4x4 enemyRotation;
    bool dead;
    int life;
    int maxlife;
    std::map<int, std::map<std::string, animElement> > animMap;
    int frameNum;
    int frameMax;
    std::string name;
    Uint32 lastShotTimer;
};

#endif

