#include "scene.hpp"
#include <iostream>

int SceneNode::last_id = 1;
std::map<std::string, std::map<int, SceneNode*> >SceneNode::nodeMap;
std::map<std::string, std::map<std::string, SceneNode*> > SceneNode::nodeNameMap;


SceneNode::SceneNode(const std::string& name, const std::string& mastername)
  : m_name(name)
{
  m_master = mastername;
  m_id = last_id;
  last_id++;
  nodeMap[mastername][m_id] = this;
  nodeNameMap[mastername][name] = this;
  parent = NULL;
}

SceneNode::~SceneNode()
{
}

void SceneNode::walk_gl(double alpha) const
{
  ChildList::const_iterator iter;
  
  glPushMatrix();
  
  glMultMatrixd(m_trans.transpose().begin());  
  
  for (iter=m_children.begin(); iter != m_children.end(); iter++)
  {
    (*iter)->walk_gl(alpha);
  }
  glPopMatrix();
  
}

void SceneNode::rotate(char axis, double angle)
{
  //std::cerr << "Stub: Rotate " << m_name << " around " << axis << " by " << angle << std::endl;
  if (angle == 0) return;
  
  Matrix4x4 r;  //this is identity
  angle = angle * (M_PI / (double) 180);
  if (axis == 'x')
  { 
    r[1][1] = cos(angle);
    r[1][2] = -sin(angle);
    r[2][1] = sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'y')
  { 
    r[0][0] = cos(angle);
    r[0][2] = sin(angle);
    r[2][0] = -sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'z')
  { 
    r[0][0] = cos(angle);
    r[0][1] = -sin(angle);
    r[1][0] = sin(angle);
    r[1][1] = cos(angle);

  }

   m_trans = m_trans * r;
  
  m_invtrans = m_trans.invert();  
}

void SceneNode::scale(const Vector3D& amount)
{
  //std::cerr << "Stub: Scale " << m_name << " by " << amount << std::endl;
  // Fill me in
  Matrix4x4 s;
  for (int i = 0; i < 3; i++)
    s[i][i] = amount[i];
  m_trans = m_trans * s;
  m_invtrans = m_trans.invert();  
}

void SceneNode::translate(const Vector3D& amount)
{
  //std::cerr << "Stub: Translate " << m_name << " by " << amount << std::endl;
  // Fill me in
  Matrix4x4 t;
  for (int i = 0; i < 3; i++)
    t[i][3] = amount[i]; 
  m_trans = m_trans * t; 
  m_invtrans = m_trans.invert();  
}

bool SceneNode::is_joint() const
{
  return false;
}

SceneNode* SceneNode::get_joint()
{
  SceneNode *cur = parent;
  while (true)
  {
    if (cur == NULL) return NULL;
    if (cur->is_joint()) return cur;
    cur = cur->parent;
  }
}


JointNode::JointNode(const std::string& name,const std::string& mastername )
  : SceneNode(name, mastername)
{
  m_dispJoint = false;
}

JointNode::~JointNode()
{
}

Matrix4x4 JointNode::rotateRet(char axis, double angle)
{
  //std::cerr << "Stub: Rotate " << m_name << " around " << axis << " by " << angle << std::endl;
  
  Matrix4x4 r;  //this is identity
  if (angle == 0) return r;
  angle = angle * (M_PI / (double) 180);
  if (axis == 'x')
  { 
    r[1][1] = cos(angle);
    r[1][2] = -sin(angle);
    r[2][1] = sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'y')
  { 
    r[0][0] = cos(angle);
    r[0][2] = sin(angle);
    r[2][0] = -sin(angle);
    r[2][2] = cos(angle);
  }

  if (axis == 'z')
  { 
    r[0][0] = cos(angle);
    r[0][1] = -sin(angle);
    r[1][0] = sin(angle);
    r[1][1] = cos(angle);

  }

   return r;
}


void JointNode::walk_gl(double alpha) const
{
  ChildList::const_iterator iter;
  
  glPushMatrix();
  
  
  Matrix4x4 trans = m_trans * m_joint_rot;
  
  glMultMatrixd(trans.transpose().begin());  
  
  
  
  for (iter=m_children.begin(); iter != m_children.end(); iter++)
  {
    (*iter)->walk_gl(alpha);
  }
  glPopMatrix();

}

bool JointNode::is_joint() const
{
  return true;
}

void JointNode::set_joint_x(double min, double init, double max)
{
  m_joint_x.min = min;
  m_joint_x.init = init;
  m_joint_x.max = max;
  m_joint_x.cur = init;
  
  m_joint_rot = rotateRet('y', m_joint_y.cur) * rotateRet('x', m_joint_x.cur); 
}


void JointNode::set_joint_y(double min, double init, double max)
{
  m_joint_y.min = min;
  m_joint_y.init = init;
  m_joint_y.max = max;
  m_joint_y.cur = init;

  m_joint_rot = rotateRet('y', m_joint_y.cur) * rotateRet('x', m_joint_x.cur); 
}

void JointNode::set_joint_x(double angle)
{

  m_joint_x.cur = angle;
  if (m_joint_x.cur > m_joint_x.max) m_joint_x.cur = m_joint_x.max;
  if (m_joint_x.cur < m_joint_x.min) m_joint_x.cur = m_joint_x.min;

  m_joint_rot = rotateRet('y', m_joint_y.cur) * rotateRet('x', m_joint_x.cur); 
}

void JointNode::set_joint_y(double angle)
{
  m_joint_y.cur = angle;
  if (m_joint_y.cur > m_joint_y.max) m_joint_y.cur = m_joint_y.max;
  if (m_joint_y.cur < m_joint_y.min) m_joint_y.cur = m_joint_y.min;

  m_joint_rot = rotateRet('y', m_joint_y.cur) * rotateRet('x', m_joint_x.cur); 
}

void JointNode::reset_joints()
{
  m_joint_x.cur = m_joint_x.init; 
  m_joint_y.cur = m_joint_y.init;
  m_joint_rot = rotateRet('y', m_joint_y.cur) * rotateRet('x', m_joint_x.cur); 
}

GeometryNode::GeometryNode(const std::string& name, const std::string& mastername, Primitive* primitive)
  : SceneNode(name, mastername),
    m_primitive(primitive)
{
}

GeometryNode::~GeometryNode()
{
}

void GeometryNode::walk_gl(double alpha) const
{


  //leaf node
  
  glPushMatrix();
  glMultMatrixd(m_trans.transpose().begin());  

    
      if (m_material != 0)
        m_material->apply_gl(alpha);    
      if (m_primitive != 0)
        m_primitive->walk_gl();
    
     
  glPopMatrix();
  
  
}






 
