#include "viewer.hpp"


std::vector<shotObj> Viewer::shots;
std::vector<shotObj> Viewer::shotSecondary;



//http://www.gamedev.net/reference/articles/article1234.asp
bool sphereTest(Point3D pos1, Vector3D dir1, double rad1, Point3D pos2, Vector3D dir2, double rad2)
{
  Vector3D    dv    = dir2 - dir1;
  Vector3D    dp    = pos2 - pos1;
  
  double r = rad1 + rad2;
  
  
  double sphereDist = dp.dot(dp) - r*r; 
  
  if ( sphereDist < 0 ) return true;

  
  double moveAway = dp.dot(dv); 
  
  if ( moveAway >= 0 ) return false;

  
  double quickIntersect = dv.dot(dv);
  
  if ( (moveAway + quickIntersect) <= 0 && (quickIntersect + 2 * moveAway + sphereDist) >= 0 ) return false;

  double mindist = -moveAway/quickIntersect;

  //Discriminant/(4*dV^2) = -(dp^2-r^2+dP*dV*tmin)
  return ( sphereDist + moveAway * mindist > 0 );
}

Matrix4x4 lookAt(Vector3D dir, Vector3D up) 
{ 
  Vector3D z = dir; 
  z.normalize();
  Vector3D x = up.cross(z);
  x.normalize(); 
  Vector3D y = z.cross(x);
  Matrix4x4 m;

  m[0][0] = x[0];
  m[0][1] = x[1];
  m[0][2] = x[2];
  m[1][0] = y[0];
  m[1][1] = y[1];
  m[1][2] = y[2];
  m[2][0] = z[0];
  m[2][1] = z[1];
  m[2][2] = z[2];

  return m.transpose();
} 
    

void decelerate(Vector3D &v, double factor)
{
    if (v[2] > 0)
      v[2] = v[2] - factor * v[2];
    if (v[2] < 0)
      v[2] = v[2] - factor * v[2];
    if (v[2] > -0.01 && v[2] < 0.01) v[2] = 0;

    if (v[0] > 0)
      v[0] = v[0] - factor * v[0];
    if (v[0] < 0)
      v[0] = v[0] - factor * v[0];
    if (v[0] > -0.01 && v[0] < 0.01) v[0] = 0;
    

    if (v[1] > 0)
      v[1] = v[1] - factor * v[1];
    if (v[1] < 0)
      v[1] = v[1] - factor * v[1];
    if (v[1] > -0.01 && v[1] < 0.01) v[1] = 0;
    

}


template<typename T> std::string toString(T x) 
{
  std::ostringstream ss;
  ss << x;
  return ss.str();
}


int intFromString(const std::string &s)
{
 std::istringstream iss(s);
 int t;
 iss >> t;
 return t;
}

double doubleFromString(const std::string &s)
{
 std::istringstream iss(s);
 double t;
 iss >> t;
 return t;
}



void Viewer::loadTexture(GLuint holder, std::string imageName, double numElem)
{
    Image textureMap;
    textureMap.loadPng(imageName);
    glBindTexture(GL_TEXTURE_2D, holder);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    if (numElem == 3)
      glTexImage2D(GL_TEXTURE_2D, 0, 3, textureMap.width(), textureMap.height(), 0, GL_RGB, GL_UNSIGNED_BYTE, textureMap.realdata());
    else if (numElem == 4)
      glTexImage2D(GL_TEXTURE_2D, 0, 4, textureMap.width(), textureMap.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMap.realdata());
} 

Viewer::Viewer(int fullsc)
{
  showTexture = true;
  godView = false;
  godMode = false;

  musicEnabled = true;
  soundEnabled = true;
  frustumCull = false;

  
  
  timeWaster = 0;
  
  helpText.push_back(TextLine("Playing Zygote:", 0, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("---------------", 1, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("CURSOR UP/DOWN         - Adjust Pitch", 2, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("CURSOR LEFT/RIGHT      -   Adjust Yaw", 3, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("CURSOR DELETE/PAGEDOWN -  Adjust Roll", 4, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("SPACE                  -         Fire", 5, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("Q, W                   -    Flagulate", 6, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("~                      -      Console", 7, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  helpText.push_back(TextLine("ESC                    -         Quit", 8, Colour(1,1,1), TextLine::PLACE_ALIGN, this));
  
  
  
  
  
  
  
  
  
  
  
  
  tailtick = 0;


  showHelp = false;
  
  windowWidth = 640;
  windowHeight = 480;


  fullscreen = fullsc;


  if(CreateGLWindow("SDL & OpenGL", windowWidth, windowHeight, 32, fullscreen) == 0){
      printf("Could not initalize OpenGL :(\n\n");
      KillGLWindow();
      exit(-1);
  }



  if (fullscreen) SDL_ShowCursor(0);
  
  alternateInput = false;
  

  animFlag = false;
  animCount = 0;



  godViewX = 0;
  godViewY = 0;
  godViewZ = 0;

  oldx = 0;
  oldy = 0;

 
  filter = 2;

  music_on = 0;

  rotational_amt = 1;

  wireframe = false;
  fog = true;

  debug = false;

  fpsCount = 0;
  fpsTimer = 0;
  fps = 0;

 //camtrans = translation(Vector3D(0,0,-100));

    
  int done=0; 

  
  

  

  Sphere::init();
  spermRoot = import_lua("./model/sperm.lua", "SPERM");
  loadKeys("SPERM", "./model/sperm.key");
  frameNumMap["SPERM"] = 0;
  
  
  loadKeys("SPERM2", "./model/sperm.key");
  frameNumMap["SPERM2"] = 0;
  SceneNode *spermRoot2 = import_lua("./model/sperm.lua", "SPERM2");
  enemyList.push_back(new Enemy("SPERM2", spermRoot2, animMap["SPERM2"], 79, Point3D(-135.969, -130.687, 5.93586), 30));  
  
  
  loadKeys("SPERM3", "./model/sperm.key");
  SceneNode *spermRoot3 = import_lua("./model/sperm.lua", "SPERM3");
  frameNumMap["SPERM3"] = 0;
  enemyList.push_back(new Enemy("SPERM3", spermRoot3, animMap["SPERM3"], 79, Point3D(-386.444, 65.3916, 23.3831), 5));      
    
  loadKeys("SPERM4", "./model/sperm.key");
  SceneNode *spermRoot4 = import_lua("./model/sperm.lua", "SPERM4");
  frameNumMap["SPERM4"] = 0;
  enemyList.push_back(new Enemy("SPERM4", spermRoot4, animMap["SPERM4"], 79, Point3D(-432.232, 443.881, 22.0099), 45));      
    
        
  loadKeys("SPERM5", "./model/sperm.key");
  SceneNode *spermRoot5 = import_lua("./model/sperm.lua", "SPERM5");
  frameNumMap["SPERM5"] = 0;
  enemyList.push_back(new Enemy("SPERM5", spermRoot5, animMap["SPERM5"], 79, Point3D(178.681, -95.2716, 1.14158), 1));      

  loadKeys("SPERM6", "./model/sperm.key");
  SceneNode *spermRoot6 = import_lua("./model/sperm.lua", "SPERM6");
  frameNumMap["SPERM6"] = 0;
  enemyList.push_back(new Enemy("SPERM6", spermRoot6, animMap["SPERM6"], 79, Point3D(328.874, -384.293, 8.71588), 100));      
  
  loadKeys("SPERM7", "./model/sperm.key");
  SceneNode *spermRoot7 = import_lua("./model/sperm.lua", "SPERM7");
  frameNumMap["SPERM7"] = 0;
  enemyList.push_back(new Enemy("SPERM7", spermRoot7, animMap["SPERM7"], 79, Point3D(-289.243, 479.278, 7.26021), 75));      
  
  loadKeys("SPERM8", "./model/sperm.key");
  SceneNode *spermRoot8 = import_lua("./model/sperm.lua", "SPERM8");
  frameNumMap["SPERM8"] = 0;
  enemyList.push_back(new Enemy("SPERM8", spermRoot8, animMap["SPERM8"], 79, Point3D(-482.866, -506.727, -15.2009), 42));      







    
    
    
    
    glGenTextures(1, &texture[0]);    
    Image textureMap;
    textureMap.loadPng("./texture/text1.png");
    glBindTexture(GL_TEXTURE_2D, texture[0]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, 3, textureMap.width(), textureMap.height(), 0, GL_RGB, GL_UNSIGNED_BYTE, textureMap.realdata());

    
    glGenTextures(1, &texture[1]);
    glBindTexture(GL_TEXTURE_2D, texture[1]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, 3, textureMap.width(), textureMap.height(), 0, GL_RGB, GL_UNSIGNED_BYTE, textureMap.realdata());

    glGenTextures(1, &texture[2]);
    glBindTexture(GL_TEXTURE_2D, texture[2]);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, textureMap.width(), textureMap.height(), GL_RGB, GL_UNSIGNED_BYTE, textureMap.realdata());

 
    glGenTextures(1, &texture[3]);
    loadTexture(texture[3], "./texture/font.png", 3);
    
    glGenTextures(1, &texture[4]);
    loadTexture(texture[4], "./texture/console.png", 4);

    glGenTextures(1, &texture[5]);
    loadTexture(texture[5], "./texture/particle3.png", 3);

    glGenTextures(1, &texture[6]);
    loadTexture(texture[6], "./texture/chromosomebak.png", 3);
    
    glGenTextures(1, &texture[7]);
    loadTexture(texture[7], "./texture/particlehit.png", 3);
    
    glGenTextures(1, &texture[8]);
    loadTexture(texture[8], "./texture/particlehit2.png", 3);
    
    glGenTextures(1, &texture[9]);
    loadTexture(texture[9], "./texture/ovum.png", 3);
    
    glGenTextures(1, &texture[10]);
    loadTexture(texture[10], "./texture/tissue.png", 3);

    glGenTextures(1, &texture[11]);
    loadTexture(texture[11], "./texture/zygote.png", 4);

    glGenTextures(1, &texture[12]);
    loadTexture(texture[12], "./texture/indicator.png", 4);

    glGenTextures(1, &texture[13]);
    loadTexture(texture[13], "./texture/bluefill.png", 3);

    glGenTextures(1, &texture[14]);
    loadTexture(texture[14], "./texture/greenfill.png", 3);

    glGenTextures(1, &texture[15]);
    loadTexture(texture[15], "./texture/radar.png", 3);



  int changeit = 1;

  makeMickey();
  
  

  env = new World("./texture/disp2.png", 15, scaling(Vector3D(20,20,20)));
  env2 = new World("./texture/disp3.png", -15, translation(Vector3D(0,0,20*15)) * rotation(degToRad(180), 'x') * scaling(Vector3D(20,20,20)));
  BuildFont();
  
  
  
  
  
  
  
  
  
  
  
  gameConsole.setFontTexture(texture[3], texture[4], fontBase);
  gameConsole.writeConsole(ConsoleItem("Zygote", CommentType));
  gameConsole.writeConsole(ConsoleItem("------", CommentType));
  gameConsole.writeConsole(ConsoleItem("", CommentType));
  gameConsole.writeConsole(ConsoleItem("Created By: Shawn Henderson", CommentType));
  gameConsole.writeConsole(ConsoleItem("Final Project: CS488 at the University of Waterloo", CommentType));
  gameConsole.writeConsole(ConsoleItem("", CommentType));
  gameConsole.writeConsole(ConsoleItem(std::string("Game Version: 0.1.0"), CommentType));
  gameConsole.writeConsole(ConsoleItem(std::string("SDL Version: ") + toString(SDL_MAJOR_VERSION) + "." + toString(SDL_MINOR_VERSION) + "." + toString(SDL_PATCHLEVEL), CommentType));
  gameConsole.writeConsole(ConsoleItem(std::string("OpenGL Version: ") + std::string((const char*)glGetString(GL_VERSION)), CommentType));

  camtrans = translation(Vector3D(0,0,-10)) * rotation(degToRad(90), 'x') * translation(Vector3D(0,0,10)) * camtrans;
  camtransInv = camtrans.invert();
  
  orientation[0] = orientation[0] + 90;



  soundMap["CHARHITWALL"] = SM.LoadSound("./sound/charhitwall.wav");
  soundMap["CHARDIE"] = SM.LoadSound("./sound/chardie.wav");
  soundMap["SHOTHITWALL"] = SM.LoadSound("./sound/shothitwall.wav");
  soundMap["SHOTHITCHAR1"] = SM.LoadSound("./sound/shothitchar1.wav");
  soundMap["SHOTHITCHAR2"] = SM.LoadSound("./sound/shothitchar2.wav");
  soundMap["SHOTHITCHAR3"] = SM.LoadSound("./sound/shothitchar3.wav");
  soundMap["CHARHITCHAR"] = SM.LoadSound("./sound/charhitchar.wav");
  soundMap["POWERUP"] = SM.LoadSound("./sound/powerup.wav");
  soundMap["SHOOT"] = SM.LoadSound("./sound/shoot.wav");
  soundMap["TITLE"] = SM.LoadMusic("./sound/title.mp3");
  soundMap["GAME"] = SM.LoadMusic("./sound/wanda.mp3");
  soundMap["CREDITS"] = SM.LoadMusic("./sound/credits.mp3");




}









void Viewer::displayIndicators()
{

      glMatrixMode( GL_PROJECTION );
      glPushMatrix( );

        glLoadIdentity( );
        glOrtho( 0, 640, 0, 480, -1, 1 );

        glMatrixMode( GL_MODELVIEW );
        glPushMatrix( );
          glLoadIdentity( );

	  
          glPushMatrix();
         
            glDisable( GL_DEPTH_TEST );
            glDisable( GL_LIGHTING );
            glEnable(GL_BLEND);
    
            
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	  
            glEnable(GL_TEXTURE_2D);

            glBindTexture( GL_TEXTURE_2D, texture[13] );
	        glColor3d(1,1,1);
            
 
            glPushMatrix();
              glTranslated(0, 60, 0);

              glBegin( GL_QUADS );
     
                glTexCoord2f( 0+indicatorTextureOffset, (double)38/256 );
                glVertex2i( 1, 38 );

                glTexCoord2f( (double)19/256+indicatorTextureOffset, (double)38/256 );
                glVertex2i( 19, 38 );

                glTexCoord2f( (double)19/256+indicatorTextureOffset, 0 );    
                glVertex2i( 19, 38+207*((double)lifeAmtSlow/100) );
    
    
                glTexCoord2f( 0+indicatorTextureOffset, 0 );
                glVertex2i( 1, 38+207*((double)lifeAmtSlow/100) );
    
             glEnd();


              glBindTexture( GL_TEXTURE_2D, texture[12] );
	  
              glColor3d(1,1,1);
              glBegin( GL_QUADS );
   
                glTexCoord2f( 0, (double)275/512 );
                glVertex2i( 0, 0 );

                glTexCoord2f( (double)20/64, (double)275/512 );
                glVertex2i( 20, 0 );

                glTexCoord2f( (double)20/64, 0 );    
                glVertex2i( 20, 275 );
    
    
                glTexCoord2f( 0, 0 );
                glVertex2i( 0, 275 );
    
             glEnd();
            glPopMatrix();

            glPushMatrix();
            glTranslated(620, 60, 0);
            glBindTexture( GL_TEXTURE_2D, texture[14] );
	        glColor3d(1,1,1);
            
              glBegin( GL_QUADS );
     
                glTexCoord2f( 0+indicatorTextureOffset, (double)38/256 );
                glVertex2i( 1, 38 );

                glTexCoord2f( (double)19/256+indicatorTextureOffset, (double)38/256 );
                glVertex2i( 19, 38 );

                glTexCoord2f( (double)19/256+indicatorTextureOffset, 0 );    
                glVertex2i( 19, 38+207*((double)ammoAmtSlow/100) );
    
    
                glTexCoord2f( 0+indicatorTextureOffset, 0 );
                glVertex2i( 1, 38+207*((double)ammoAmtSlow/100) );
    
             glEnd();


            glBindTexture( GL_TEXTURE_2D, texture[12] );
	  
	        glColor3d(1,1,1);
            glBegin( GL_QUADS );
   
              glTexCoord2f( 0, (double)275/512 );
              glVertex2i( 0, 0 );

              glTexCoord2f( (double)20/64, (double)275/512 );
              glVertex2i( 20, 0 );

              glTexCoord2f( (double)20/64, 0 );    
              glVertex2i( 20, 275 );
    
    
              glTexCoord2f( 0, 0 );
              glVertex2i( 0, 275 );
    
           glEnd();
           glPopMatrix();

	   
	   /*

           glBlendFunc(GL_SRC_ALPHA, GL_ONE);
           glBindTexture( GL_TEXTURE_2D, texture[15] );
	  
           glColor3d(1,1,1);
           glBegin( GL_QUADS );
   
             glTexCoord2f( 0, 0 );
             glVertex2i( 0, 0 );

             glTexCoord2f(1 , 0 );
             glVertex2i( 50, 0 );

             glTexCoord2f(1,1 );    
             glVertex2i( 50, 50 );
    
    
             glTexCoord2f( 0, 1 );
             glVertex2i( 0, 50 );
    
           glEnd();
*/



          glPopMatrix();
          glDisable(GL_TEXTURE_2D);
	      glDisable(GL_BLEND);
          glEnable( GL_LIGHTING );
          glEnable( GL_DEPTH_TEST );


          glDisable(GL_TEXTURE_2D);
          glMatrixMode( GL_PROJECTION );
          glPopMatrix( );

      glMatrixMode( GL_MODELVIEW );
      glPopMatrix( );
    glPopMatrix();


}













bool Viewer::run()
{
  

  SDL_Delay(1000);
  
  envCollide = true;
  
  glClearColor(0.20f, 0.00f, 0.00f, 0.0f);
  
  
  
  indicatorTextureOffset = 0;

  bool done = false;
  int changeit = 1;
  Uint8* keys; 

  gameConsole.writeConsoleDisp(ConsoleItem("Hold F1 for Help", InfoType));

  
  spermAlpha = 1;
  alive = true;
  ammoAmt = 100;
  lifeAmt = 100;
  ammoRegenRate = 0.1;
  ammoAmtSlow = ammoAmt;
  lifeAmtSlow = lifeAmt;

  
  
  
    if (musicEnabled)
    {
      SM.StopMusic(soundMap["TITLE"]);
      SM.PlayMusic(soundMap["GAME"], true);
    }


  
  
  
  
  
  
  
  while(!done){

  /*
  float projection[16];
  int i,j;

  glGetFloatv(GL_PROJECTION_MATRIX , projection);
  Matrix4x4 proj;
  for( i=0; i<4; i++ ) 
   for( j=0; j<4; j++ ) 
   {
      proj[i][j] = projection[i*4+j];
   }

  
  Matrix4x4 tot =  camtrans; 
  cam.loadFrustum(tot.transpose());
*/
  
  
  if (timeWaster > 0)
    SDL_Delay(timeWaster);



   if (godMode) lifeAmt = 100;
   if (godMode) ammoAmt = 100;

   if (lifeAmt <= 0 && alive)
   {      
     shotSecondary.push_back(shotObj(curpos, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 5000, 0.7, 0.2, Colour(1,0,0), 0.3, ParticleEngine::PARTICLE_EXPLODE, texture[6]));
     alive = false;
     velocity = Vector3D(0,0,0);
     rotVelocity = Vector3D(0,0,0);
     if (soundEnabled) SM.PlaySound(soundMap["CHARDIE"]);
     deathTime = SDL_GetTicks() + 5000; 
  }

  
  //bail out after a sec
  if (!alive)
  {
    if (SDL_GetTicks() > deathTime) done = true;
  }








     

    //need to get the normal of the triangle we just hit
    Vector3D normal;
    double dist;
    Vector3D vel = camtransInv*velocity; //camtrans * translation(Vector3D(0,0,10)) * velocity;
    double magnitude = vel.normalize();
    if (env->intersect(curpos, vel, 2, false, normal, dist) && envCollide)
    {
     if (soundEnabled) SM.PlaySound(soundMap["CHARHITWALL"]);

     //std::cerr << "INTERSECTION1" << "   " << normal << "   " << dist << "  " << vel << std::endl;
     //if (dist < 0) { normal = -1*normal; }
     normal = camtrans * (-1 * normal);
     normal.normalize();
     double distToPush = (1+dist);
     if (distToPush > 1) distToPush = 0.1;
     camtrans = translation(distToPush*normal) * camtrans;
     camtransInv = camtrans.invert();
     //velocity = 0.3 * normal;
     
     //velocity = 0.05*normal; //Vector3D(0,0,0);
     velocity = Vector3D(0,0,0);
     
     //Vector3D r = (-1*vel) + (2.0 * (vel.dot(normal))) * normal;
     //velocity = camtrans * (magnitude*0.8*r);
    }
    
    if (env2->intersect(curpos, vel, 2, true, normal, dist) && envCollide)
    {
     if (soundEnabled) SM.PlaySound(soundMap["CHARHITWALL"]);

     //std::cerr << "INTERSECTION2" << "   " << normal << "   " << dist << "  " << vel << std::endl;
     //if (dist < 0) { normal = -1*normal; }
     normal = camtrans * (-1 * normal);
     normal.normalize();
     double distToPush = (1-dist);
     if (distToPush > 1) distToPush = 0.1;

     camtrans = translation(distToPush*normal) * camtrans;
     camtransInv = camtrans.invert();
     //velocity = 0.3 * normal;
     
     //velocity = 0.05*normal; //Vector3D(0,0,0);
     velocity = Vector3D(0,0,0);
    }


    
    
    
    
    
    
    
    
    
    
    
         

    //need to get the normal of the triangle we just hit
    for (int i = 0; i < enemyList.size(); i++)
    {
      if (enemyList[i]->alive())
      {
        vel = enemyList[i]->getFullVel();
        if (env->intersect(enemyList[i]->getPos(), vel, 2, true, normal, dist))
        {
          //Spare us hearing that crap.
	  //if (soundEnabled) SM.PlaySound(soundMap["CHARHITWALL"]);

          //std::cerr << "Enemy INTERSECTION1" << "   " << normal << "   " << dist << "  " << vel << std::endl;
          //if (dist < 0) { normal = -1*normal; }
          normal = camtrans * (-1 * normal);
          normal.normalize();
          double distToPush = (1+dist);
          if (distToPush > 1) distToPush = 0.5;
          enemyList[i]->setPos(enemyList[i]->getPos() + (distToPush*normal));
          enemyList[i]->setVel(-0.3*enemyList[i]->getVel());
        }
    
        if (env2->intersect(enemyList[i]->getPos(), vel, 2, false, normal, dist))
        {
          //if (soundEnabled) SM.PlaySound(soundMap["CHARHITWALL"]);

          //std::cerr << "Enemy INTERSECTION2" << "   " << normal << "   " << dist << "  " << vel << std::endl;
          //if (dist < 0) { normal = -1*normal; }
          normal = camtrans * (-1 * normal);
          normal.normalize();
          double distToPush = (1-dist);
          if (distToPush > 1) distToPush = 0.1;

          enemyList[i]->setPos(enemyList[i]->getPos() + (distToPush*normal));
          enemyList[i]->setVel(-0.3*enemyList[i]->getVel());
        }
      }
    }
    
    
    
    
    
    
    
    
    
    
    
    


  
    std::string consoleCommand = gameConsole.getLastCommand();
    if (!consoleCommand.empty())
    {
      runCommand(consoleCommand);
    }
    // Draw the scene
    // And poll for events
    if (gameConsole.state() != ConsoleDown)
    {
      handleSmoothInput();
    }
    
    
      
    //update indicators
    if (ammoAmt > ammoAmtSlow)
      ammoAmtSlow = ammoAmtSlow + 0.8;
    if (ammoAmt < ammoAmtSlow)
      ammoAmtSlow = ammoAmtSlow - 0.8;
    if (lifeAmt > lifeAmtSlow)
      lifeAmtSlow = lifeAmtSlow + 0.8;
    if (lifeAmt < lifeAmtSlow)
      lifeAmtSlow = lifeAmtSlow - 0.8;




  
    //inertia
    camtrans =  camtrans * translation(camtrans.invert()*velocity);
   
   
    
    decelerate(velocity, 0.02);
    decelerate(rotVelocity, 0.05);
        
    camtrans = translation(Vector3D(0,0,-10)) * rotation(degToRad(rotVelocity[0]), 'x') * translation(Vector3D(0,0,10))* camtrans;
    camtrans = translation(Vector3D(0,0,-10)) * rotation(degToRad(rotVelocity[1]), 'y') * translation(Vector3D(0,0,10))* camtrans;
    camtrans = translation(Vector3D(0,0,-10)) * rotation(degToRad(rotVelocity[2]), 'z') * translation(Vector3D(0,0,10))* camtrans;
   
    camtransInv = camtrans.invert();
    orientation = orientation + rotVelocity;
    
    
    
    if (orientation[0] < 0) orientation[0] = orientation[0] + 360;
    if (orientation[0] > 360) orientation[0] = orientation[0] - 360;
    if (orientation[1] < 0) orientation[1] = orientation[1] + 360;
    if (orientation[1] > 360) orientation[1] = orientation[1] - 360;
    if (orientation[2] < 0) orientation[2] = orientation[2] + 360;
    if (orientation[2] > 360) orientation[2] = orientation[2] - 360;    
    
    
    
    
    

    
    for (int i = 0; i < enemyList.size(); i++)
    {
      if (enemyList[i]->alive())
      {
        enemyList[i]->updateOrientation(curpos-enemyList[i]->getPos(), Vector3D(0,1,0));
        EnemyActivity res = enemyList[i]->tick(curpos);
   
        //sh0ot at me
        if (res == ENEMY_SHOOT)
        {
           if (soundEnabled) SM.PlaySound(soundMap["SHOOT"]);
	       Vector3D tome = (curpos - enemyList[i]->getPos()) + (-1*(camtransInv*velocity));
           tome.normalize();
           Point3D shotpos = enemyList[i]->getPos();
           Vector3D shotmove = 2*tome;
           Vector3D shotacc = -0.008 * shotmove;
      
           shots.push_back(shotObj(shotpos, shotmove, shotacc, SHOT_OWNER_ENEMYSPERM, 10, 500, 0.7, 0.2, Colour(0,1,0), 0.3, ParticleEngine::PARTICLE_FIRE, texture[5]));
        }
	if (res == ENEMY_JUST_DIED || res == ENEMY_DROP_PRESENT) //newly dead
	{
           shotSecondary.push_back(shotObj(enemyList[i]->getPos(), Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 3000, 0.7, 0.2, Colour(1,0,0), 0.3, ParticleEngine::PARTICLE_EXPLODE, texture[6]));	
	}
        if (res == ENEMY_DROP_PRESENT)
	{
           powerups.push_back(pickupObj(PICKUP_GLUCOSE, 1000, enemyList[i]->getPos(), 250, 0.2, 0.7, Colour(1,1,1), 0.3, ParticleEngine::PARTICLE_FIRE, texture[5]));
	} 
      }
    }
   
   
   
   
   
   
   
   ammoAmt = ammoAmt + ammoRegenRate;
   if (ammoAmt > 100) ammoAmt = 100;
    
    
    
   std::vector< std::vector<shotObj>::iterator > killList;
   for (std::vector<shotObj>::iterator it = shots.begin(); it != shots.end(); it++)
   {
     if ((*it).shotdone)
     {

       killList.push_back(it);
       
     }  
     else  
     {  
       Vector3D normal;
       double dist;
       if ((*it).pmode != ParticleEngine::PARTICLE_EXPLODE && (*it).collidable)
       {
	 for (int i = 0; i < enemyList.size(); i++)
	 {
	   if (sphereTest((*it).shotpos, Vector3D(0,0,0), 1, enemyList[i]->getPos(), Vector3D(0,0,0), 1) && (*it).owner != SHOT_OWNER_ENEMYSPERM && enemyList[i]->alive())
	   {
             if (soundEnabled) 
             { 
               int num = rand()%3;
               if (num == 0) SM.PlaySound(soundMap["SHOTHITCHAR1"]);
               if (num == 1) SM.PlaySound(soundMap["SHOTHITCHAR2"]);
               if (num == 2) SM.PlaySound(soundMap["SHOTHITCHAR3"]);
             }
           
             shotSecondary.push_back(shotObj((*it).shotpos, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 50, 0.2, 0.8, Colour(0.8,0.3,0.0), 0, ParticleEngine::PARTICLE_EXPLODE, texture[8]));
             gameConsole.writeConsoleDisp(ConsoleItem("Enemy Hit", InfoType));
	     //enemyVelocity = enemyVelocity + enemyRotation.invert()*((-0.1)*(*it).shotmove); 
	     (*it).kill();
	     //std::cerr << "ENEMYVEL: " << enemyRotation*((-0.1)*(*it).shotmove) << std::endl;
	     enemyList[i]->doDamage((*it).damage);
	  }
        }
	if (sphereTest((*it).shotpos, Vector3D(0,0,0), 1, curpos, Vector3D(0,0,0), 1) && (*it).owner != SHOT_OWNER_ME && alive)
	 {
           if (soundEnabled) 
           {
               int num = rand()%3;
               if (num == 0) SM.PlaySound(soundMap["SHOTHITCHAR1"]);
               if (num == 1) SM.PlaySound(soundMap["SHOTHITCHAR2"]);
               if (num == 2) SM.PlaySound(soundMap["SHOTHITCHAR3"]);
           }
           
	   //normal = camtrans * (-2 * normal);
           //normal.normalize();
 	       velocity = velocity + camtrans*((-0.1)*(*it).shotmove); 

           (*it).kill();
           shotSecondary.push_back(shotObj((*it).shotpos, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 50, 0.2, 0.8, Colour(0.8,0.3,0.0), 0, ParticleEngine::PARTICLE_EXPLODE, texture[8]));
           gameConsole.writeConsoleDisp(ConsoleItem("Membrane Damage", InfoType));
	 	   lifeAmt = lifeAmt - (*it).damage;
           if (lifeAmt < 0) lifeAmt = 0;

       } 

	 else if (env->intersect((*it).shotpos, (*it).shotmove, 0, true, normal, dist) ||  env2->intersect((*it).shotpos, (*it).shotmove, 0, false, normal, dist)) 
         {

           if (soundEnabled) SM.PlaySound(soundMap["SHOTHITWALL"]);

           normal.normalize();
           //normal = camtrans * (1 * normal);
	       (*it).kill();
           shotSecondary.push_back(shotObj((*it).shotpos + normal, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 250, 0.2, 0.8, Colour(0.1,0.8,0.1), 0, ParticleEngine::PARTICLE_EXPLODE, texture[7]));
           shotSecondary.push_back(shotObj((*it).shotpos + normal, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 250, 0.2, 0.8, Colour(0.1,0.8,0.1), 0, ParticleEngine::PARTICLE_EXPLODE, texture[8]));

           gameConsole.writeConsoleDisp(ConsoleItem("Shot Hit Wall", InfoType));
         }
	 
       }

     }  
     (*it).tick();    
   }  
    
   
   for (int i = 0; i < killList.size(); i++)
     shots.erase(killList[i]);
   
   killList.clear();
   for (std::vector<shotObj>::iterator it = shotSecondary.begin(); it != shotSecondary.end(); it++)
   {
     if ((*it).shotdone)
     {
       killList.push_back(it);
       
     }  
     else  
       (*it).tick();    
   }  
   
   for (int i = 0; i < killList.size(); i++)
     shotSecondary.erase(killList[i]);
   
   
     
     
     

     
     
     
     
   std::vector< std::vector<pickupObj>::iterator > powerKillList;
   for (std::vector<pickupObj>::iterator it = powerups.begin(); it != powerups.end(); it++)
   {
     if ((*it).done)
     {
       powerKillList.push_back(it);
       
     }  
     else  
     {  
       Vector3D normal;
       double dist;
       if (sphereTest((*it).objpos, Vector3D(0,0,0), 1, curpos, Vector3D(0,0,0), 1) && (*it).collidable)
       {
           (*it).kill();
           gameConsole.writeConsoleDisp(ConsoleItem("Got Glucose", InfoType));
           if (soundEnabled) SM.PlaySound(soundMap["POWERUP"]);
           lifeAmt = lifeAmt + 20; 
           if (lifeAmt > 100) lifeAmt = 100;           
       }

     }  
     (*it).tick();    
   }  
        
   for (int i = 0; i < powerKillList.size(); i++)
     powerups.erase(powerKillList[i]);
     
     
 
     
 //enemycollision    
 {

       Vector3D normal;
       double dist;
       for (int i = 0; i < enemyList.size(); i++)
       {
         if (sphereTest(curpos, Vector3D(0,0,0), 0.5, enemyList[i]->getPos(), Vector3D(0,0,0), 0.5) &&enemyList[i]->alive())
         {
             enemyList[i]->kill();
             gameConsole.writeConsoleDisp(ConsoleItem("Collision", InfoType));
             lifeAmt = lifeAmt - 50;
             if (lifeAmt < 0) lifeAmt = 0;
             shotSecondary.push_back(shotObj(curpos, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 1000, 0.7, 0.2, Colour(1,0,0), 0.3, ParticleEngine::PARTICLE_EXPLODE, texture[6]));
         }
       }

}




     
   indicatorTextureOffset = indicatorTextureOffset + 0.001;
   if (indicatorTextureOffset > 1) indicatorTextureOffset = 0;
     
     
     

   if (!alive)
   {
     spermAlpha = spermAlpha - 0.05;
     if (spermAlpha < 0) spermAlpha = 0;  
   }
     

    if (changeit == 1)
       {
       DrawGLScene();
       changeit = 1;
       }

            
                
    SDL_Event event;
    while ( SDL_PollEvent(&event) ) {
      switch (event.type) {
      case SDL_KEYDOWN:
      case SDL_KEYUP:
	  handleKey(event.key);
      if (event.key.keysym.sym == SDLK_ESCAPE) done = true;
	  changeit = 1;
	  break;

      
        
      case SDL_MOUSEMOTION:
	  //SDL_PeepEvents (&event,9,SDL_GETEVENT,SDL_MOUSEMOTION);
	  handleMouseMotion(event);
      //this affects the screen.
      changeit = 1;
	  break;


      case SDL_MOUSEBUTTONDOWN:
      case SDL_MOUSEBUTTONUP:
	  handleMouseButtons(event.button);
	  break;
        case SDL_QUIT:
          // then we're done and we'll end this program
          return false;
          done = 1;
          break;
        default:
          break;
      }// switch

    }// while 

    // Get the state of the keyboard keys
    keys = SDL_GetKeyState(NULL);

    
    
    
   if (SDL_GetTicks() > tailtick)
   {
    animTick("SPERM"); 
    tailtick = SDL_GetTicks() + 25;
   } 
   for (int i = 0; i < (int)(velocity[2]*1.5); i++) 
     animTick("SPERM"); 
    
    
    
  }// while done


  if (musicEnabled)
  {
    SM.StopMusic(soundMap["GAME"]);
  }


 return true;

  
}






void Viewer::BuildFont()
{
    GLuint loop;
    double cx;
    double cy;

    fontBase  = glGenLists( 256 );
    glBindTexture( GL_TEXTURE_2D, texture[3] );

    glPushMatrix();
    for ( loop = 0; loop < 256; loop++ )
    {
	    cx = 1 - ( double )( loop % 16 ) / 16.0f;
	    cy = 1 - ( double )( loop / 16 ) / 16.0f;

	    glNewList( fontBase + ( 255 - loop ), GL_COMPILE );
	      glBegin( GL_QUADS );
	        glTexCoord2f( cx - 0.0625, cy );
            glVertex2i( 0, 0 );

            glTexCoord2f( cx, cy );
            glVertex2i( 16, 0 );

            glTexCoord2f( cx, cy - 0.0625f );
            glVertex2i( 16, 16 );

            glTexCoord2f( cx - 0.0625f, cy - 0.0625f);
            glVertex2i( 0, 16 );
          glEnd( );
	      glTranslated( 10, 0, 0 );
	    glEndList( );
  }
  glPopMatrix();
}





GLvoid Viewer::glPrint( GLint x, GLint y, char *string, double scalef)
{
    glEnable(GL_TEXTURE_2D);
    glBindTexture( GL_TEXTURE_2D, texture[3] );

    glDisable( GL_DEPTH_TEST );
    glDisable( GL_LIGHTING );
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);

    glMatrixMode( GL_PROJECTION );
    glPushMatrix( );

      glLoadIdentity( );
      glOrtho( 0, 640, 0, 480, -1, 1 );

      glMatrixMode( GL_MODELVIEW );
      glPushMatrix( );
        glLoadIdentity( );

        glTranslated( x, y, 0 );
    
        glScaled(scalef, scalef, scalef);
    
        glListBase( fontBase - 32);

        glCallLists( strlen( string ), GL_BYTE, string );
    
        glMatrixMode( GL_PROJECTION );
      glPopMatrix( );

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix( );

    glDisable(GL_BLEND);
    glEnable( GL_LIGHTING );
    glEnable( GL_DEPTH_TEST );
    glDisable(GL_TEXTURE_2D);
}







void Viewer::makeMickey()
{
        GLUquadricObj* qsphere = gluNewQuadric();
	glNewList(MICKEY, GL_COMPILE);
        
    	gluQuadricDrawStyle(qsphere, GLU_FILL);
    	gluQuadricNormals(qsphere, GLU_SMOOTH);
    	gluQuadricOrientation(qsphere, GLU_OUTSIDE);
    	gluQuadricTexture(qsphere, GL_FALSE);
        glPushMatrix();
    	glColor3f(1,1,0);
    	gluSphere(qsphere, 13, 20, 20);
    	glTranslatef(14,14,0);
    	gluSphere(qsphere, 10, 20, 20);
    	glTranslatef(-28,0,0);
    	gluSphere(qsphere, 10, 20, 20);
    	glPopMatrix();
	glEndList();
        
	glNewList(SPHERE, GL_COMPILE);
	gluQuadricTexture(qsphere, GL_TRUE);
	gluSphere(qsphere, 1, 16, 16);
	
	glEndList();
	
	gluDeleteQuadric(qsphere);


glNewList(PARTICLE_LIST, GL_COMPILE);
    glBegin(GL_TRIANGLE_STRIP);
      glNormal3d(0,0,1); glTexCoord2d(1,1); glVertex3f(1,1,0); // Top Right
      glNormal3d(0,0,1); glTexCoord2d(0,1); glVertex3f(-1,1,0); // Top Left
      glNormal3d(0,0,1); glTexCoord2d(1,0); glVertex3f(1,-1,0); // Bottom Right
      glNormal3d(0,0,1); glTexCoord2d(0,0); glVertex3f(-1,-1,0); // Bottom Left      
    glEnd();
glEndList();



}


	


void Viewer::DrawGLScene(){
 


   Matrix4x4 proj, mod;
   double mat[16];

   glGetDoublev(GL_PROJECTION_MATRIX , mat);
   proj = Matrix4x4(mat);
    
   mod = camtrans;
   
   Matrix4x4 tot =  mod;//*proj; //*proj.transpose();
   cam.loadFrustum(tot);

 
  
 
  if (wireframe)
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  else
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
   
  
  
  
  if (fog)
    glEnable(GL_FOG);
  else
    glDisable(GL_FOG);   
  
  glEnable ( GL_COLOR_MATERIAL ) ;
  glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;
  glShadeModel(GL_SMOOTH);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);


//  glEnable(GL_RESCALE_NORMAL);
  glEnable(GL_NORMALIZE);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  
  glMatrixMode(GL_MODELVIEW);
    
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glLoadIdentity(); // Reset the view

    
    

    
    
    
  glPushMatrix();

    // draw s simple sphere
    


  
    if (godView)
    {
      glTranslated(godViewX, godViewY, godViewZ);
      glMultMatrixd(godViewRotate.transpose().begin());
    }
    glMultMatrixd(camtrans.transpose().begin());
    

    
    Cam *camPointer = &cam;
    if (!frustumCull) camPointer = NULL;

    
    if (!wireframe)
    {
      if (showTexture) 
      {
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, texture[filter]);
      }
      env->render(camPointer, curtime);
      env2->render(camPointer, curtime);
      if (showTexture) 
      {
        glDisable(GL_TEXTURE_2D);
      }
   }
   else
   {
      glDisable(GL_TEXTURE_2D);
      glColor3d(1,1,1);
      env->render(camPointer, curtime);
      env2->render(camPointer, curtime);


   } 
    
    
    
            
    glPushMatrix();
    
      glMultMatrixd(camtransInv.transpose().begin());
      glTranslated(0,0,-10);
          
       //set up the light 
       GLfloat specular [] = { 0.9, 0.9, 0.9, 1 };
       GLfloat shininess [] = { 20.0 };
       GLfloat ambient[] = {0.2, 0.2, 0.2, 1};
       GLfloat diffuse[] = {0.8, 0.8, 0.8, 1};
       GLfloat position1 [] = { 0.0, 10.0, 5.0, 1.0 };
  
       glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
       glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

       glLightfv(GL_LIGHT0, GL_POSITION, position1);      
       glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);      
       glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);      
    
       glEnable(GL_LIGHT0);
       glEnable(GL_LIGHTING);
      

       
      
      glPushMatrix();
        glRotated(100, 1,0,0);
        glRotated(90, 0,1,0);	
        glScaled(0.4, 0.4, 0.4); 
        
	if (spermAlpha < 1)
	{
          glDepthMask(GL_FALSE);
          glEnable(GL_BLEND);
          glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
        }
	  spermRoot->walk_gl(spermAlpha);
	if (spermAlpha < 1)
	{
	  glDepthMask(GL_TRUE);
	  glDisable(GL_BLEND);
	}
      glPopMatrix();


    glPopMatrix();


   

  
  
{

    for (int i = 0; i < enemyList.size(); i++)
      if (enemyList[i]->alive())
         enemyList[i]->render();
    
    


}

      








        
    
    
    curpos = camtransInv * translation(Vector3D(0,0,-10)) * Point3D(0,0,0);
    curvel = velocity;
    //curpos = getmat * camtrans.invert() * translation(Vector3D(0,0,-10)) * Point3D(0,0,0);
    //curvel = getmat * camtrans.invert() * translation(Vector3D(0,0,-10)) * velocity;

/*  
      
    glPushMatrix();
    glTranslated(curpos[0], curpos[1], curpos[2]);
    GLUquadricObj* qsphere = gluNewQuadric();
    glColor3d(1,0,0);
    gluSphere(qsphere, 1, 5, 5);
    gluDeleteQuadric(qsphere);
    glPopMatrix();
*/

    


  glDisable(GL_FOG);    
  
  for (int i = 0; i < shots.size(); i++)
    shots[i].render();
  for (int i = 0; i < shotSecondary.size(); i++)
    shotSecondary[i].render();
  for (int i = 0; i < powerups.size(); i++)
    powerups[i].render();


  displayIndicators();

     
  if (fog)
    glEnable(GL_FOG);      
    

glPopMatrix();
    
  
  
    curtime = SDL_GetTicks();
    fpsCount = fpsCount + 1;
    if (curtime > fpsTimer)
    {
      fps = fpsCount;
      fpsCount = 0;
      fpsTimer = curtime + 1000;
    }
    
    if (debug)
    
    {
    std::ostringstream strout;
    strout << std::setprecision(4);
    
    strout << "POS: " << curpos;
    std::string tstr = strout.str();
    strout.str("");
    
    glEnable(GL_TEXTURE_2D);
    glColor3d(1,1,1);
    glPrint(0, 440, const_cast<char*>(tstr.c_str()), 1);
    glDisable(GL_TEXTURE_2D);

    strout << "VEL: " << velocity;
    tstr = strout.str();
    strout.str("");
    
    glEnable(GL_TEXTURE_2D);
    glColor3d(1,1,1);
    glPrint(0, 440-(1*1*16), const_cast<char*>(tstr.c_str()), 1);
    glDisable(GL_TEXTURE_2D);
    
    
    strout << "ANG: " << orientation;
    tstr = strout.str();
    strout.str("");
    
    glEnable(GL_TEXTURE_2D);
    glColor3d(1,1,1);
    glPrint(0, 440-(2*1*16), const_cast<char*>(tstr.c_str()), 1);
    glDisable(GL_TEXTURE_2D);
    
    
    strout << "FPS: " << fps;
    tstr = strout.str();
    strout.str("");
    
    glEnable(GL_TEXTURE_2D);
    glColor3d(1,1,1);
    glPrint(0, 440-(3*1*16), const_cast<char*>(tstr.c_str()), 1);
    glDisable(GL_TEXTURE_2D);
   
    strout << "ROT: " << rotVelocity;
    tstr = strout.str();
    strout.str("");
    
    glEnable(GL_TEXTURE_2D);
    glColor3d(1,1,1);
    glPrint(0, 440-(4*1*16), const_cast<char*>(tstr.c_str()), 1);
    glDisable(GL_TEXTURE_2D);

    }
            
    
    
   
    glEnable(GL_TEXTURE_2D);
    gameConsole.drawConsole();
    glDisable(GL_TEXTURE_2D);

    if (showHelp)
    {
      for (int i = 0; i < helpText.size(); i++)
      {
        helpText[i].print();
      }
    }
    
 /*
   for (int i = 0; i < enemyList.size(); i++)
   {
    Vector3D tome = (curpos - enemyList[i]->getPos()) + (-1*(camtransInv*velocity));
    //tome.normalize();
    glBegin(GL_LINES);
      glVertex3dv(enemyList[i]->getPos().begin());
      glVertex3dv((enemyList[i]->getPos() + 10*tome).begin());
    glEnd();

   }
*/   

 SDL_GL_SwapBuffers(); // Swap the buffers
    
    
    
}


void Viewer::handleKey(SDL_KeyboardEvent key) {
  
  if (gameConsole.state() != ConsoleUp)
  {
    gameConsole.update(key);
    switch(key.keysym.sym) {
      case SDLK_BACKQUOTE:
        if(key.state == SDL_PRESSED) 
        gameConsole.toggle();
        break;
    }
  }  
  else
  {
    switch(key.keysym.sym) {
  
  
    case SDLK_z:
      if(key.state == SDL_PRESSED) {
        wireframe = !wireframe;
        if (wireframe)
       	  gameConsole.writeConsoleDisp(ConsoleItem("Wireframe On", InfoType));
        else
       	  gameConsole.writeConsoleDisp(ConsoleItem("Wireframe Off", InfoType));
      }// if
      break;
    /*  
    case SDLK_f:
    if(key.state == SDL_PRESSED) 
    {
      if (filter == 0) 
       runCommand("filter linear");
      else if (filter == 1)
       runCommand("filter mipmap");
      else
       runCommand("filter nearest");
    }
      break;
    case SDLK_g:
    if(key.state == SDL_PRESSED)
    { 
      if (fog) 
       runCommand("fog off");
      else
       runCommand("fog on");
    }
      break;
    */
  

    case SDLK_BACKQUOTE:
    if(key.state == SDL_PRESSED) 
      gameConsole.toggle();
      break;

     
    case SDLK_SPACE:
    if(key.state == SDL_PRESSED) 
    {
      if (alive && ammoAmt > 33)
      {
        ammoAmt = ammoAmt - 33;
        if (soundEnabled) SM.PlaySound(soundMap["SHOOT"]);
        Point3D shotpos = curpos;// + camtrans.invert() * translation(Vector3D(0,0,-10)); // * (-1*Vector3D(0,0,5));
        Vector3D shotmove = camtransInv * translation(Vector3D(0,0,-10)) * (-1*(velocity+Vector3D(0,0,2)));
        Vector3D shotacc = -0.01 * shotmove;
      
        shots.push_back(shotObj(shotpos, shotmove, shotacc, SHOT_OWNER_ME, 20, 100, 0.2, 0.8, Colour(0,1,0), 0, ParticleEngine::PARTICLE_FIRE, texture[5]));
        velocity[2] = velocity[2] - 0.3;
      }
      else
        gameConsole.writeConsoleDisp(ConsoleItem("Insufficient Enzyme", InfoType));
    }
      break;


      
         

      
         

         
   case SDLK_w:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{
	  if (alternateInput == false)
	  {
	    if (velocity[2] < 0.05) 
	      velocity = velocity + Vector3D(0,0,0.05);
	    else
	      velocity[2] = velocity[2] + 0.5 * velocity[2];
	    if (velocity[2] > 3) velocity[2] = 3;
            alternateInput = true;
	  }
	  
	}      
      }// if
      break;
        
   case SDLK_q:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{ 
	  if (alternateInput == true)
	  {
	    if (velocity[2] < 0.05) 
	      velocity = velocity + Vector3D(0,0,0.05);
	    else
	      velocity[2] = velocity[2] + 0.5 * velocity[2];
	    if (velocity[2] > 3) velocity[2] = 3;

            alternateInput = false;
	  }
	  
	}      
      }// if
      break;

   // case SDLK_1:
   //   if(key.state == SDL_PRESSED) {
   //     std::cerr << "Point3D(" << curpos[0] << ", " << curpos[1] << ", " << curpos[2] << ")" << std::endl;
   //   }// if
   //   break;
   /*   
      
   case SDLK_1:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{
	    velocity = velocity + Vector3D(1,0,0);
	}      
      }// if
      break;
      
      
   case SDLK_2:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{
	    velocity = velocity + Vector3D(-1,0,0);
	}      
      }// if
      break;
      
   case SDLK_3:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{
	    velocity = velocity + Vector3D(0,1,0);
	}      
      }// if
      break;
      
   case SDLK_4:
      if(key.state == SDL_PRESSED) {
        if (alive)
	{
	    velocity = velocity + Vector3D(0,-1,0);
	}      
      }// if
      break;
      
   */
      
      
      
        
  
    }// switch
  }
}



void Viewer::handleMouseMotion(SDL_Event event){
    
    SDL_MouseButtonEvent button = event.button;


    if ((button.button & SDL_BUTTON(3)) && godView)
    {
      int current_width = windowWidth;
      int current_height = windowHeight;  
      double radius = current_width < current_height ?  (float)current_width * 0.25 : (float)current_height * 0.25;
      Vector3D track;
      Matrix4x4 rot;
  
      int iCenterX = current_width / 2;
      int iCenterY = current_height / 2;
      double vnewX = event.motion.x - iCenterX;
      double vnewY = event.motion.y - iCenterY;
      double voldX = oldx - iCenterX;
      double voldY = oldy - iCenterY;
          
      vCalcRotVec(vnewX, vnewY, voldX, voldY, radius*2, track);
      track[1] = -track[1];
      vAxisRotMatrix(track, rot);
  

      godViewRotate = rot * godViewRotate;
      
      
    }

    if ((button.button &SDL_BUTTON(1)) && godView)
    {
      godViewX = godViewX + double(event.motion.x-oldx)/10.0;
      godViewY = godViewY - double(event.motion.y-oldy)/10.0;

    }
    if ((button.button &SDL_BUTTON(2)) && godView)
    {
      godViewZ = godViewZ + double(event.motion.y-oldy)/10.0;
    }

    oldx = event.motion.x;
    oldy = event.motion.y;
    
    
    
    
    
      

    
    
    
    
    
    
}



inline void Viewer::handleMouseButtons(SDL_MouseButtonEvent button){
    
    if(button.type == SDL_MOUSEBUTTONDOWN){
	switch (button.button){
	case 1/*LMB*/: break;
	case 2/*MMB*/: break;
	case 3/*RMB*/: break;
	}// switch
    } else { /* SDL_MOUSEBUTTONUP */
	switch (button.button){
	case 1: break;
	case 2: break;
	case 3: break;
	} // switch
    }// if ... else

}



std::vector<std::string> Viewer::explode(std::string str, std::string delimiters = " ")
{

    std::vector<std::string> tokens;
    int lastPos = str.find_first_not_of(delimiters, 0);
    int pos     = str.find_first_of(delimiters, lastPos);

    while (std::string::npos != pos || std::string::npos != lastPos)
    {
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        lastPos = str.find_first_not_of(delimiters, pos);
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}


void Viewer::runCommand(std::string cmd)
{
  cmd.erase(cmd.find_last_not_of(" ")+1);
  cmd.erase(0, cmd.find_first_not_of (" "));
  std::vector<std::string> cmdline = explode(cmd, " ");
  
  gameConsole.writeConsole(ConsoleItem(cmd, CommandType));
  if (cmdline.size() > 0)
  {
    
    if (cmdline[0] == "list")
    {
       gameConsole.writeConsole(ConsoleItem("List of Commands",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  debug [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  fog [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  playsound [slot]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  filter [nearest|linear|mipmap]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  showtexture [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  shownormal [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  showfacenormal [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  showbound [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  animatenormal [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  setenvcollide [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  music [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  sound [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  frustumcull [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  godmode [on|off]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  godview [on|off|reset]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  setfarplane [num]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  setdelay [num]",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  die",  InfoType));
       gameConsole.writeConsole(ConsoleItem("  quit",  InfoType));    
    }
    
    else if (cmdline[0] == "debug")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
	{
	  debug = true;
	  gameConsole.writeConsoleDisp(ConsoleItem("Debug Mode: On", InfoType));
	}
        else if (cmdline[1] == "off")
	{
	  debug = false;
	  gameConsole.writeConsoleDisp(ConsoleItem("Debug Mode: Off", InfoType));
	}
        else
	{
	  gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
	}

      }
    }
    else if (cmdline[0] == "fog")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          fog = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Fog On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          fog = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Fog Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }

    else if (cmdline[0] == "playsound")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        int slot = intFromString(cmdline[1]);
      	SM.PlaySound(slot);

      }
    }



    else if (cmdline[0] == "filter")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "nearest")
	{
	  filter = 0;
	  gameConsole.writeConsoleDisp(ConsoleItem("Filter Nearest", InfoType));
	}
        else if (cmdline[1] == "linear")
	{
	  filter = 1;
	  gameConsole.writeConsoleDisp(ConsoleItem("Filter Linear", InfoType));
	}
        else if (cmdline[1] == "mipmap")
	{
	  filter = 2;
	  gameConsole.writeConsoleDisp(ConsoleItem("Filter mipmap", InfoType));
	}
	else
	{
	  gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
	}

      }
    }
    
   
   
   
   
   else if (cmdline[0] == "showtexture")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          showTexture = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Texture On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          showTexture = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Texture Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }    
    
    else if (cmdline[0] == "shownormal")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          World::showNormal = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Normal On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          World::showNormal = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Normal Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }

    else if (cmdline[0] == "showfacenormal")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          World::showFaceNormal = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Face Normal On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          World::showFaceNormal = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Face Normal Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
    
    
    
    else if (cmdline[0] == "showbound")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          World::showBound = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Bound On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          World::showBound = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Show Bound Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
    
    
    else if (cmdline[0] == "animatenormal")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          World::animateNormal = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Animate Normal On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          World::animateNormal = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Animate Normal Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }

    
    else if (cmdline[0] == "setenvcollide")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          envCollide = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Environment Collisions On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          envCollide = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Environment Collisions off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
    
    else if (cmdline[0] == "music")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          musicEnabled = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Music On", InfoType));
          SM.PlayMusic(soundMap["TITLE"], true);
        }
        else if (cmdline[1] == "off")
        { 
          musicEnabled = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Music Off", InfoType));
          SM.StopMusic(soundMap["TITLE"]);

        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
   else if (cmdline[0] == "sound")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          soundEnabled = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Sound Effects On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          soundEnabled = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Sound Effects Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
        
   else if (cmdline[0] == "frustumcull")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          frustumCull = true;
          gameConsole.writeConsoleDisp(ConsoleItem("Frustum Culling On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          frustumCull = false;
          gameConsole.writeConsoleDisp(ConsoleItem("Frustum Culling Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
  
   else if (cmdline[0] == "godmode")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
           godMode = true;
           gameConsole.writeConsoleDisp(ConsoleItem("God Mode On", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
           godMode = false;
           gameConsole.writeConsoleDisp(ConsoleItem("God Mode Off", InfoType));
        }
        else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
   
   else if (cmdline[0] == "godview")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        if (cmdline[1] == "on")
        {
          godView = true;
          SDL_ShowCursor(1);

          gameConsole.writeConsoleDisp(ConsoleItem("Godview Enabled", InfoType));
        }
        else if (cmdline[1] == "off")
        { 
          godView = false;
	  if (fullscreen) SDL_ShowCursor(0);
          gameConsole.writeConsoleDisp(ConsoleItem("Godview Disabled", InfoType));
        }
        else if (cmdline[1] == "reset")
        { 
          godViewX = 0;
	  godViewY = 0;
	  godViewZ = 0;
	  godViewRotate = Matrix4x4();
          gameConsole.writeConsoleDisp(ConsoleItem("Environment Collisions off", InfoType));
        }
        
	else
        {
          gameConsole.writeConsole(ConsoleItem("Error: Unexpected parameter value", ErrorType));
        }

      }
    }
    
        
    
    
    
    
    else if (cmdline[0] == "setfarplane")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        double newfar = doubleFromString(cmdline[1]);
        glMatrixMode(GL_PROJECTION);					
        glLoadIdentity();									

        // Calculate The Aspect Ratio Of The Window
        gluPerspective(40.0, (GLfloat)windowWidth/(GLfloat)windowHeight, 0.1, newfar);

        glMatrixMode(GL_MODELVIEW);		

      }
    }
    
    else if (cmdline[0] == "setdelay")
    {
      if (cmdline.size() < 2)
      {
        gameConsole.writeConsole(ConsoleItem("Error: Expected parameter", ErrorType));
      }
      else
      {
        timeWaster = intFromString(cmdline[1]);
        if (timeWaster < 0) timeWaster = 0; 

      }
    }    
    
        
    

    else if (cmdline[0] == "quit")
    {
      SDL_Event qe;
      qe.type = SDL_QUIT;
      SDL_PushEvent(&qe);
    }
    
    else if (cmdline[0] == "die")
    {
        shotSecondary.push_back(shotObj(curpos, Vector3D(0,0,0), Vector3D(0,0,0), SHOT_OWNER_NOBODY, 0, 5000, 0.7, 0.2, Colour(1,0,0), 0.3, ParticleEngine::PARTICLE_EXPLODE, texture[6]));
     alive = false;
     velocity = Vector3D(0,0,0);
     rotVelocity = Vector3D(0,0,0);
     if (soundEnabled) SM.PlaySound(soundMap["CHARDIE"]);
     deathTime = SDL_GetTicks() + 5000; 
    } 
    
    
    else
    {
      gameConsole.writeConsole(ConsoleItem("Error: Unknown command", ErrorType));
    }
  }
    
}


void Viewer::handleSmoothInput()
{
    Uint8 *keystate = SDL_GetKeyState(NULL);
    
    if (alive)
    {
      if ( keystate[SDLK_UP])
      { 
        rotVelocity[0] = rotVelocity[0] - (0.05 * rotational_amt);
        if (rotVelocity[0] < -rotational_amt) rotVelocity[0] = -rotational_amt;
      }
      if ( keystate[SDLK_DOWN]) 
      {
        rotVelocity[0] = rotVelocity[0] + (0.05 * rotational_amt);
        if (rotVelocity[0] > rotational_amt) rotVelocity[0] = rotational_amt;
      }
      if ( keystate[SDLK_LEFT])
      { 
        rotVelocity[1] = rotVelocity[1] - (0.05 * rotational_amt);
        if (rotVelocity[1] < -rotational_amt) rotVelocity[1] = -rotational_amt;
      }
      if ( keystate[SDLK_RIGHT]) 
      {
        rotVelocity[1] = rotVelocity[1] + (0.05 * rotational_amt);
        if (rotVelocity[1] > rotational_amt) rotVelocity[1] = rotational_amt;

      }
      if ( keystate[SDLK_DELETE]) 
      {
        rotVelocity[2] = rotVelocity[2] + (0.05 * rotational_amt);
        if (rotVelocity[2] > rotational_amt) rotVelocity[2] = rotational_amt;

      }
      if ( keystate[SDLK_PAGEDOWN]) 
      {
        rotVelocity[2] = rotVelocity[2] - (0.05 * rotational_amt);
        if (rotVelocity[2] < -rotational_amt) rotVelocity[2] = -rotational_amt;
      }
    
    
    
      if ( keystate[SDLK_a] )
      { 
          camtrans = translation(Vector3D(0,0,-10)) * camtrans;
          camtransInv = camtrans.invert();
      }
      if ( keystate[SDLK_s] ) 
      {  
          camtrans = translation(Vector3D(0,0,10)) * camtrans;
          camtransInv = camtrans.invert();
      }
    }
    
    
    showHelp = false;
    if ( keystate[SDLK_F1] ) 
    {
      showHelp = true;
    }
    
    
}


bool Viewer::loadKeys(std::string masterName, std::string keyfilename)
{
  std::map<int, std::map<std::string, animElement> > frameMap;
  
  std::ifstream fin(keyfilename.c_str());
  if (fin.fail()) return false;

  std::string getkeyident;
  fin >> getkeyident;
  if (getkeyident != "key") return false;
  
  int upper = 0;  
  while(true)
  {
    int keynum;
    std::string tstr;
    fin >> tstr;
    keynum = intFromString(tstr);
    if (keynum > upper) upper = keynum;

    //readChunk(fin, keynum);
    {
      std::map<std::string, animElement> keyframe;
      std::string objname;
      std::string angxstr;
      std::string angystr;
  
      while(true)
      {
        fin >> objname;
        if (objname == "key") break;
        if (fin.eof()) break;
        fin >> angxstr;
        fin >> angystr;
        animElement ae(doubleFromString(angxstr), doubleFromString(angystr));
        keyframe[objname] = ae;
      }
      if (!keyframe.empty())
        frameMap[keynum] = keyframe;
    
    }
    if (fin.eof()) break;
  }

  animMap[masterName] = frameMap;

  for (int i = 0; i <= upper; i++)
  {
   frameNumMap[masterName] = i;
   for (std::map<int, SceneNode*>::iterator it = SceneNode::nodeMap[masterName].begin(); it != SceneNode::nodeMap[masterName].end(); it++)
    {
      JointNode *jointNode = dynamic_cast<JointNode*>((*it).second);
      if (jointNode != 0)
        interpolate(masterName, jointNode->get_name());
    }

  }
  frameNumMap[masterName] = 0;

  return true;
}



void Viewer::animTick(std::string masterName)
{
  
  if (animMap[masterName].upper_bound(frameNumMap[masterName]) == animMap[masterName].end())
  {
    frameNumMap[masterName] = 0;
  }
  frameNumMap[masterName]++;
  for (std::map<int, SceneNode*>::iterator it = SceneNode::nodeMap[masterName].begin(); it != SceneNode::nodeMap[masterName].end(); it++)
  {
    JointNode *jointNode = dynamic_cast<JointNode*>((*it).second);
    if (jointNode != 0)
    {
      std::string jointName = jointNode->get_name();  
      std::map<std::string, animElement> keyframe = animMap[masterName][frameNumMap[masterName]];
      animElement *ae = &keyframe[jointName];
      jointNode->set_joint_x(ae->anglex);
      jointNode->set_joint_y(ae->angley);
    }
  }
}


//interpolate the position for the joint Jointname
void Viewer::interpolate(std::string masterName, std::string jointName)
{
  
  
  std::map<int, std::map<std::string, animElement> > *frameAnimMap = &animMap[masterName];
  
  //currently in an animation frame
  if ((*frameAnimMap).find(frameNumMap[masterName]) != (*frameAnimMap).end())
  {

    std::map<std::string, animElement> keyframe = (*frameAnimMap)[frameNumMap[masterName]];
    
    if (keyframe.find(jointName) != keyframe.end())
    {
      //do nothing      
      return;
    }
    
  }
  

  //so for this joint, we must be between nodes
  
  bool prev_exist = false;
  bool next_exist = false;
  int next_frame = 0;
  int prev_frame = 0;

  animElement prev_keyframe;
  animElement next_keyframe;
 
  
  int upperbound;
  if (animMap.empty())
    upperbound = 0;
  else
  {
    std::map<int, std::map<std::string, animElement> >::iterator it = (*frameAnimMap).end();
    it--;
    upperbound = (*it).first;
    
  }

  {
    for (int i = frameNumMap[masterName]; i <= upperbound; i++)
    {
       if ((*frameAnimMap).find(i) != (*frameAnimMap).end())
       {
         std::map<std::string, animElement> elem = (*frameAnimMap)[i];
	     if (elem.find(jointName) != elem.end())
         {
           next_exist = true;
           next_frame = i;
           next_keyframe = elem[jointName];
           break;
         }
       }
    }
  }
  
  {
    
    
    for (int i = frameNumMap[masterName]; i >= 0; i--)
    {
       if ((*frameAnimMap).find(i) != (*frameAnimMap).end())
       {
         std::map<std::string, animElement> elem = (*frameAnimMap)[i];
	 if (elem.find(jointName) != elem.end())
	 {
	   prev_exist = true;
	   prev_frame = i;
	   prev_keyframe = elem[jointName];
	   break;
	 }
       }
    }
  }




  
  //std::cerr << jointName << "   " << m_framenum << " " << prev_exist << " " << prev_frame << "  " << next_exist << " " << next_frame << " " << upperbound << std::endl; 
  
  if (prev_exist && next_exist)
  {
     
     double prev_angx = prev_keyframe.anglex;
     double prev_angy = prev_keyframe.angley;

     double next_angx = next_keyframe.anglex;
     double next_angy = next_keyframe.angley;
     double new_angx = prev_angx + ((next_angx - prev_angx) * ((double)(frameNumMap[masterName] - prev_frame)) / (next_frame - prev_frame));
     double new_angy = prev_angy + ((next_angy - prev_angy) * ((double)(frameNumMap[masterName] - prev_frame)) / (next_frame - prev_frame));

     animElement ae(new_angx, new_angy);
     animMap[masterName][frameNumMap[masterName]][jointName] = ae;


  }
  else
  {
     
     JointNode *jointNode = (JointNode*)SceneNode::nodeNameMap[masterName][jointName];

     animElement ae(jointNode->get_joint_x(), jointNode->get_joint_y());
     animMap[masterName][frameNumMap[masterName]][jointName] = ae;
  }
   

}





bool Viewer::intro()
{
  

  if (musicEnabled)
  {
    SM.PlayMusic(soundMap["TITLE"], true);
  }

  

  glClearColor(0.2f, 0.0f, 0.0f, 0.0f);
  bool done = false;
  Uint8* keys; 

  intro_doneOpenScreen = false;
  intro_openScreenAlpha = 0;
  intro_timer = 0;
  intro_fadeTimer = 0;
  intro_scrollAmt = 0;

  
  gameMenu = GAME_SELECT_START;

  

  while(true){

    DrawIntroScene();

    if (done && SDL_GetTicks() > intro_fadeTimer)
    {
      intro_openScreenAlpha = intro_openScreenAlpha +0.05;
      intro_fadeTimer = SDL_GetTicks() + 30;
    }
    
    
    
     

    
    
    
    if (intro_openScreenAlpha >= 1) break;    

    
    if (SDL_GetTicks() > intro_timer)
    {
      intro_scrollAmt = intro_scrollAmt - 0.02;
      if (intro_scrollAmt < -1) intro_scrollAmt = 0;
      intro_timer = SDL_GetTicks() + 20;
      animTick("SPERM"); 
    }
    
   
                       
    SDL_Event event;
    while ( SDL_PollEvent(&event) ) {
      switch (event.type) {
      case SDL_KEYDOWN:
      case SDL_KEYUP:
	  {
        if (event.key.state == SDL_PRESSED)
        {
          if (event.key.keysym.sym == SDLK_UP || event.key.keysym.sym == SDLK_DOWN )
          {
            if (gameMenu == GAME_SELECT_START) 
               gameMenu = GAME_SELECT_QUIT;
            else
               gameMenu = GAME_SELECT_START;
          }
          else
          {
             done = true; //key hit
             if (gameMenu == GAME_SELECT_QUIT) return false;
          }
        }
      
   
	  }
	  break;

      
        
      case SDL_MOUSEMOTION:
	  //SDL_PeepEvents (&event,9,SDL_GETEVENT,SDL_MOUSEMOTION);
	  //handleMouseMotion(event);
      //this affects the screen.
     
	  break;


      case SDL_MOUSEBUTTONDOWN:
      case SDL_MOUSEBUTTONUP:
	  //handleMouseButtons(event.button);
	  break;
        case SDL_QUIT:
          // then we're done and we'll end this program
          done=1;
          return false;
          break;
        default:
          break;
      }// switch

    

    }// while 

    
    
    
  }// while done

  if (musicEnabled)
  {
    SM.StopMusic(soundMap["TITLE"]);
  }

  if (gameMenu == GAME_SELECT_QUIT)
  {
    return false;
  }

  return true;

}


void Viewer::DrawIntroScene()
{

  glEnable(GL_FOG);   
  glEnable ( GL_COLOR_MATERIAL ) ;
  glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;
  glShadeModel(GL_SMOOTH);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);
  
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_NORMALIZE);
   
  
  
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  
  glMatrixMode(GL_MODELVIEW);
    
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glLoadIdentity(); // Reset the view

    
    glPushMatrix();

    
    GLfloat specular [] = { 0.9, 0.9, 0.9, 1 };
    GLfloat shininess [] = { 20.0 };
    GLfloat ambient[] = {0.2, 0.2, 0.2, 1};
    GLfloat diffuse[] = {0.8, 0.8, 0.8, 1};
    GLfloat position1 [] = { 0.0, 10.0, 5.0, 1.0 };
  
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

    glLightfv(GL_LIGHT0, GL_POSITION, position1);      
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);      
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);      
    
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
      
    
    
      
    GLUquadricObj* qcyl = gluNewQuadric();
    gluQuadricOrientation(qcyl, GLU_INSIDE);
    
    glColor3d(1,1,1);   
    
    glPushMatrix();
      glTranslated(0,0,-500);
      gluQuadricTexture(qcyl, GL_TRUE);
      glEnable(GL_TEXTURE_2D);
      glBindTexture( GL_TEXTURE_2D, texture[10] );
    
      glMatrixMode(GL_TEXTURE);
      glPushMatrix();
        glTranslated(0,intro_scrollAmt,intro_scrollAmt);
        glScaled(1, 20, 1);
        gluCylinder(qcyl, 2, 2, 500, 20, 20);
        gluDeleteQuadric(qcyl);  
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
      glDisable(GL_TEXTURE_2D);

    glPopMatrix();
    
    glPushMatrix();
      glColor3d(1,1,1);
      glTranslated(0,-1,-5-(7*intro_openScreenAlpha));
      glRotated(100, 1,0,0);
      glRotated(90, 0,1,0);	
      glScaled(0.2, 0.2, 0.2); 
      spermRoot->walk_gl(false);
    glPopMatrix();

    if (intro_openScreenAlpha <= 0)
    {
      if (gameMenu == GAME_SELECT_START) 
      {
        glColor3d(1, 1, 1);
        glPrint((640 - 2*10*10) / 2, 470-(14*16), "START GAME", 2);
        glColor3d(0.3, 0.3, 0.3);
        glPrint((640 - 2*10*9) / 2, 470-(16*16), "QUIT GAME", 2);
      }
      else
      {
        glColor3d(0.3, 0.3, 0.3);
        glPrint((640 - 2*10*10) / 2, 470-(14*16), "START GAME", 2);
        glColor3d(1, 1, 1);
        glPrint((640 - 2*10*9) / 2, 470-(16*16), "QUIT GAME", 2);
      }
   }
     
      
    glPopMatrix();
     
    



 
      glMatrixMode( GL_PROJECTION );
      glPushMatrix( );

        glLoadIdentity( );
        glOrtho( 0, 640, 0, 480, -1, 1 );

        glMatrixMode( GL_MODELVIEW );
        glPushMatrix( );
          glLoadIdentity( );

	  
          glPushMatrix();
         
            glDisable( GL_DEPTH_TEST );
            glDisable( GL_LIGHTING );
            glEnable(GL_BLEND);
    
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	  
            glEnable(GL_TEXTURE_2D);
            glBindTexture( GL_TEXTURE_2D, texture[11] );
	  
	       glColor3d(1,1,1);
            glBegin( GL_QUADS );
   
            glTexCoord2f( 0, 1 );
            glVertex2i( 64, 470-256 );

            glTexCoord2f( 1, 1 );
            glVertex2i( 512+64, 470-256 );

            glTexCoord2f( 1, 0 );    
            glVertex2i( 512+64, 470 );
    
    
            glTexCoord2f( 0, 0 );
            glVertex2i( 64, 470 );
    
  
          glEnd();

          glDisable(GL_TEXTURE_2D);
	  

          if (intro_openScreenAlpha > 0)
          {
            glColor4d(0,0,0,intro_openScreenAlpha);
            glBegin( GL_QUADS );
              glVertex2i( 0, 0 );
              glVertex2i( 640, 0 );
              glVertex2i( 640, 480 );
              glVertex2i( 0, 480 );
            glEnd();
          }	      
	  
	      glDisable(GL_BLEND);
          glEnable( GL_LIGHTING );
          glEnable( GL_DEPTH_TEST );


          glDisable(GL_TEXTURE_2D);
          glMatrixMode( GL_PROJECTION );
          glPopMatrix( );

      glMatrixMode( GL_MODELVIEW );
      glPopMatrix( );
    glPopMatrix();
   

    SDL_GL_SwapBuffers(); // Swap the buffers
    
 }
 
 










void Viewer::outro()
{
    SDL_Delay(1000);

    if (musicEnabled)
    {
      SM.StopMusic(soundMap["GAME"]);
      SM.PlayMusic(soundMap["CREDITS"], false);
    }

    creditText.push_back(TextLine("Zygote", 0, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Created by Shawn Henderson", 4, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Final Project: CS488 UW Winter 2005", 5, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
   
    
    
    

    creditText.push_back(TextLine("ACKNOWLEDGEMENTS", 7, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    creditText.push_back(TextLine("TEXTURES", 10, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("3D Cafe (http://3dcafe.com)", 12, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Paul Bourke (http://astronomy.swin.edu.au/~pbourke/texture)", 13, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Nehe (http://nehe.gamedev.net)", 14, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("American McGee's Alice (Electronic Arts)", 15, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    creditText.push_back(TextLine("MUSIC", 17, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Title:  Random Composition 12 (Shawn Henderson)         ", 19, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Game:    WANDA WANDA (NAMCO (Yu Miyake))                ", 20, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("             from Katamari Fortissimo Damacy            ", 21, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Credit: Katamari Love (Shigeru Matsuzaki, Yohihito Yano)", 22, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("              from Katamari Fortissimo Damacy           ", 23, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("SOUNDS", 25, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("American McGee's Alice (Electronic Arts)", 27, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    creditText.push_back(TextLine("The sperm and events depicted in this game are fictitious.", 31, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Any similarity to actual sperm, living or dead,", 32, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("is purely coincidental.", 33, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    creditText.push_back(TextLine("Dislaimer:  It is possible for this game to have", 36, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("offended some groups.  However, with all the    ", 37, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("games currently available offering new and      ", 38, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("exciting opportunity the blow someone's head    ", 39, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("off, a game like this should be positively      ", 40, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("heartwarming. We also acknowledge that many     ", 41, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("artistic licenses were taken. This game is not  ", 42, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("meant as a substitute to education.             ", 43, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("                               -- The Management", 44, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    creditText.push_back(TextLine("SPECIAL THANKS", 46, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Ian Henderson", 48, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("RicMoo", 49, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Brett Rowland", 50, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("Danielle Dmytar", 51, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));
    creditText.push_back(TextLine("& Curtis Luk", 52, Colour(1,1,1), TextLine::PLACE_ALIGN_THEN_EXACT, this));

    for (int i = 0; i < creditText.size(); i++)
    {
      creditText[i].shiftY(480);
    }


  
  glClearColor(0.2f, 0.0f, 0.0f, 0.0f);
  bool done = false;
  Uint8* keys; 

  intro_doneOpenScreen = false;
  outro_openScreenAlpha = 0;
  intro_timer = 0;
  intro_scrollAmt = 0;

  outro_splitTimer = 0;
  outro_tickTimer = 0;
  outro_ova.add(Point3D(0,0,0)); 


  done = false;
  
  while(!done){
 
    
    DrawOutroScene();
           
    /*
    if (intro_doneOpenScreen && SDL_GetTicks() > intro_timer)
    {
      intro_openScreenAlpha = intro_openScreenAlpha +0.01;
      intro_timer = SDL_GetTicks() + 100;
    }
    */
    
    
     

    
    
    
//    if (intro_openScreenAlpha >= 1) done = true;    

    
    if (SDL_GetTicks() > intro_timer)
    {
      intro_scrollAmt = intro_scrollAmt - 0.001;
      if (intro_scrollAmt < -1) intro_scrollAmt = 0;
      intro_timer = SDL_GetTicks() + 15;
  
    }
    
    
     
    if (SDL_GetTicks() > outro_splitTimer)
    {
      outro_ova.spawn();
      outro_splitTimer = -1;
    }
    
    if (SDL_GetTicks() > outro_tickTimer)
    {
      bool beforeTick = outro_ova.stillMoving();
      outro_ova.tick();
      outro_ova.tick();

      bool afterTick = outro_ova.stillMoving();
      
      if (beforeTick != afterTick)
        outro_splitTimer = SDL_GetTicks() + 10000;
      
      outro_tickTimer = SDL_GetTicks() + 90;
    
       for (int i = 0; i < creditText.size(); i++)
       {
         creditText[i].shiftY(-2);
       }
      if (creditText[creditText.size()-1].ypos <= 0)
      {
        outro_openScreenAlpha = outro_openScreenAlpha +0.05;
        if (outro_openScreenAlpha > 1) { outro_openScreenAlpha = 1, done = true; }
      }

    }
    
                       
    SDL_Event event;
    while ( SDL_PollEvent(&event) ) {
      switch (event.type) {
      case SDL_KEYDOWN:
        done=true;
        break;
      case SDL_KEYUP:
	  //handleKey(event.key);
	  //intro_doneOpenScreen = true;
	  //done=true;
	  break;

      
        
      case SDL_MOUSEMOTION:
	  //SDL_PeepEvents (&event,9,SDL_GETEVENT,SDL_MOUSEMOTION);
	  //handleMouseMotion(event);
      //this affects the screen.
     
	  break;


      case SDL_MOUSEBUTTONDOWN:
      case SDL_MOUSEBUTTONUP:
	  //handleMouseButtons(event.button);
	  break;
        case SDL_QUIT:
          // then we're done and we'll end this program
          done=1;
          break;
        default:
          break;
      }// switch

    }// while 

    // Get the state of the keyboard keys
    keys = SDL_GetKeyState(NULL);

    // and check if ESCAPE has been pressed. If so then quit
    if(keys[SDLK_ESCAPE]) done=1;
    
    
    
    
    
    
  }// while done

}


void Viewer::DrawOutroScene()
{

  glEnable(GL_FOG);   
  glEnable ( GL_COLOR_MATERIAL ) ;
  glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;
  glShadeModel(GL_SMOOTH);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);
  
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_NORMALIZE);
   
  
  
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  
  glMatrixMode(GL_MODELVIEW);
    
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glLoadIdentity(); // Reset the view

    
    glPushMatrix();

    
    GLfloat specular [] = { 0.9, 0.9, 0.9, 1 };
    GLfloat shininess [] = { 20.0 };
    GLfloat ambient[] = {0.2, 0.2, 0.2, 1};
    GLfloat diffuse[] = {0.8, 0.8, 0.8, 1};
    GLfloat position1 [] = { 0.0, 10.0, 5.0, 1.0 };
  
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

    glLightfv(GL_LIGHT0, GL_POSITION, position1);      
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);      
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);      
    
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
      
    
    
      
    GLUquadricObj* qcyl = gluNewQuadric();
    gluQuadricOrientation(qcyl, GLU_INSIDE);
    
    glColor3d(1,1,1);   
    
    glPushMatrix();
      glTranslated(0,0,-500);
      gluQuadricTexture(qcyl, GL_TRUE);
      glEnable(GL_TEXTURE_2D);
      glBindTexture( GL_TEXTURE_2D, texture[10] );
    
      glMatrixMode(GL_TEXTURE);
      glPushMatrix();
        glTranslated(0,intro_scrollAmt,intro_scrollAmt);
        glScaled(1, 20, 1);
        gluCylinder(qcyl, 2, 2, 500, 20, 20);
        gluDeleteQuadric(qcyl);  
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
      glDisable(GL_TEXTURE_2D);

    glPopMatrix();
    
     
    glEnable(GL_TEXTURE_2D);
    glBindTexture( GL_TEXTURE_2D, texture[9] );
    glDepthMask( GL_FALSE );
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    
    glPushMatrix();
      glTranslated(0,0,-3);
      glScaled(0.1, 0.1, 0.1);
      outro_ova.render();

    glPopMatrix();
    
    glDisable(GL_TEXTURE_2D);
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
    
      
    glPopMatrix();

    for (int i = 0; i < creditText.size(); i++)
    {
      creditText[i].print();
    }
    

    if (outro_openScreenAlpha > 0)
    {
      glMatrixMode( GL_PROJECTION );
        glPushMatrix( );

          glLoadIdentity( );
          glOrtho( 0, 640, 0, 480, -1, 1 );

          glMatrixMode( GL_MODELVIEW );
          glPushMatrix( );
            glLoadIdentity( );

	  
            glPushMatrix();
         
              glDisable( GL_DEPTH_TEST );
              glDisable( GL_LIGHTING );
              glEnable(GL_BLEND);
    
              glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	          glDisable(GL_TEXTURE_2D);
 	  

      
              glColor4d(0,0,0,outro_openScreenAlpha);
              glBegin( GL_QUADS );
                glVertex2i( 0, 0 );
                glVertex2i( 640, 0 );
                glVertex2i( 640, 480 );
                glVertex2i( 0, 480 );
              glEnd();
  
              glDisable(GL_BLEND);
              glEnable( GL_LIGHTING );
              glEnable( GL_DEPTH_TEST );


              glDisable(GL_TEXTURE_2D);
              glMatrixMode( GL_PROJECTION );
              glPopMatrix( );

         glMatrixMode( GL_MODELVIEW );
         glPopMatrix( );
      glPopMatrix();

  }


 
    SDL_GL_SwapBuffers(); // Swap the buffers
    
 }




 
 
 
 

/*******************************************************
 * 
 * void vCalcRotVec(float fNewX, float fNewY, 
 *                  float fOldX, float fOldY,
 *                  float fDiameter,
 *                  float *fVecX, float *fVecY, float *fVecZ);
 *
 *    Calculates a rotation vector based on mouse motion over
 *    a virtual trackball.
 *
 *    The fNew and fOld mouse positions
 *    should be in 'trackball' space. That is, they have been
 *    transformed into a coordinate system centered at the middle
 *    of the trackball. fNew, fOld, and fDiameter must all be specified
 *    in the same units (pixels for example).
 *
 * Parameters: fNewX, fNewY - New mouse position in trackball space.
 *                            This is the second point along direction
 *                            of rotation.
 *             fOldX, fOldY - Old mouse position in trackball space.
 *                            This is the first point along direction
 *                            of rotation.
 *             fDiameter - Diameter of the trackball. This should
 *                         be specified in the same units as fNew and fOld.
 *                         (ie, usually pixels if fNew and fOld are transformed
 *                         mouse positions)
 *             fVec - The output rotation vector. The length of the vector
 *                    is proportional to the angle of rotation.
 *
 *******************************************************/
void Viewer::vCalcRotVec(double fNewX, double fNewY, double fOldX, double fOldY, double fDiameter, Vector3D &vec) {
   //long  nXOrigin, nYOrigin;
   double fNewVecX, fNewVecY, fNewVecZ,        /* Vector corresponding to new mouse location */
         fOldVecX, fOldVecY, fOldVecZ,        /* Vector corresponding to old mouse location */
         fLength;

   /* Vector pointing from center of virtual trackball to
    * new mouse position
    */
   fNewVecX    = fNewX * 2.0 / fDiameter;
   fNewVecY    = fNewY * 2.0 / fDiameter;
   fNewVecZ    = (1.0 - fNewVecX * fNewVecX - fNewVecY * fNewVecY);

   /* If the Z component is less than 0, the mouse point
    * falls outside of the trackball which is interpreted
    * as rotation about the Z axis.
    */
   if (fNewVecZ < 0.0) {
      fLength = sqrt(1.0 - fNewVecZ);
      fNewVecZ  = 0.0;
      fNewVecX /= fLength;
      fNewVecY /= fLength;
   } else {
      fNewVecZ = sqrt(fNewVecZ);
   }

   /* Vector pointing from center of virtual trackball to
    * old mouse position
    */
   fOldVecX    = fOldX * 2.0 / fDiameter;
   fOldVecY    = fOldY * 2.0 / fDiameter;
   fOldVecZ    = (1.0 - fOldVecX * fOldVecX - fOldVecY * fOldVecY);
 
   /* If the Z component is less than 0, the mouse point
    * falls outside of the trackball which is interpreted
    * as rotation about the Z axis.
    */
   if (fOldVecZ < 0.0) {
      fLength = sqrt(1.0 - fOldVecZ);
      fOldVecZ  = 0.0;
      fOldVecX /= fLength;
      fOldVecY /= fLength;
   } else {
      fOldVecZ = sqrt(fOldVecZ);
   }

   /* Generate rotation vector by calculating cross product:
    * 
    * fOldVec x fNewVec.
    * 
    * The rotation vector is the axis of rotation
    * and is non-unit length since the length of a crossproduct
    * is related to the angle between fOldVec and fNewVec which we need
    * in order to perform the rotation.
    */
   vec = Vector3D(fOldVecY * fNewVecZ - fNewVecY * fOldVecZ, fOldVecZ * fNewVecX - fNewVecZ * fOldVecX, fOldVecX * fNewVecY - fNewVecX * fOldVecY);
}

/*******************************************************
 * void vAxisRotMatrix(float fVecX, float fVecY, float fVecZ, Matrix mNewMat)
 *    
 *    Calculate the rotation matrix for rotation about an arbitrary axis.
 *    
 *    The axis of rotation is specified by (fVecX,fVecY,fVecZ). The length
 *    of the vector is the amount to rotate by.
 *
 * Parameters: fVecX,fVecY,fVecZ - Axis of rotation
 *             mNewMat - Output matrix of rotation in column-major format
 *                       (ie, translation components are in column 3 on rows
 *                       0,1, and 2).
 *
 *******************************************************/
void Viewer::vAxisRotMatrix(Vector3D vec, Matrix4x4 &mNewMat) {
    
    double fRadians, fInvLength, fNewVecX, fNewVecY, fNewVecZ;

    /* Find the length of the vector which is the angle of rotation
     * (in radians)
     */
    double fVecX = vec[0];
    double fVecY = vec[1];
    double fVecZ = vec[2];
    
    fRadians = sqrt(fVecX * fVecX + fVecY * fVecY + fVecZ * fVecZ);

    /* If the vector has zero length - return the identity matrix */
    if (fRadians > -0.000001 && fRadians < 0.000001) {
        mNewMat = Matrix4x4();
        return;
    }

    /* Normalize the rotation vector now in preparation for making
     * rotation matrix. 
     */
    fInvLength = 1 / fRadians;
    fNewVecX   = fVecX * fInvLength;
    fNewVecY   = fVecY * fInvLength;
    fNewVecZ   = fVecZ * fInvLength;

    /* Create the arbitrary axis rotation matrix */
    double dSinAlpha = sin(fRadians);
    double dCosAlpha = cos(fRadians);
    double dT = 1 - dCosAlpha;

    mNewMat[0][0] = dCosAlpha + fNewVecX*fNewVecX*dT;
    mNewMat[0][1] = fNewVecX*fNewVecY*dT + fNewVecZ*dSinAlpha;
    mNewMat[0][2] = fNewVecX*fNewVecZ*dT - fNewVecY*dSinAlpha;
    mNewMat[0][3] = 0;

    mNewMat[1][0] = fNewVecX*fNewVecY*dT - dSinAlpha*fNewVecZ;
    mNewMat[1][1] = dCosAlpha + fNewVecY*fNewVecY*dT;
    mNewMat[1][2] = fNewVecY*fNewVecZ*dT + dSinAlpha*fNewVecX;
    mNewMat[1][3] = 0;

    mNewMat[2][0] = fNewVecZ*fNewVecX*dT + dSinAlpha*fNewVecY;
    mNewMat[2][1] = fNewVecZ*fNewVecY*dT - dSinAlpha*fNewVecX;
    mNewMat[2][2] = dCosAlpha + fNewVecZ*fNewVecZ*dT;
    mNewMat[2][3] = 0;

    mNewMat[3][0] = 0;
    mNewMat[3][1] = 0;
    mNewMat[3][2] = 0;
    mNewMat[3][3] = 1;
}
 
 
 
 
 
