#ifndef XIX_VIEWER_HPP
#define XIX_VIEWER_HPP

#include "SDL_OGL.h" 
#include "SoundManager.h"
#include "image.hpp"
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include "algebra.hpp"
#include <vector>
#include "a2.hpp"
#include <sstream>
#include <string>
#include <iostream>
#include <iomanip>
#include "Console.hpp"
#include "scene_lua.hpp"
#include "scene.hpp"
#include "world.hpp"
#include "cam.hpp"
#include "entity.hpp"

#define MICKEY 10
#define SPHERE 11
#define QUAD 12
#define PARTICLE_LIST 13

struct TextLine;
class Enemy;







 enum GameSelectMenu 
 {
   GAME_SELECT_START,
   GAME_SELECT_QUIT
 };



 struct animElement {
   
    double anglex, angley;
    animElement() { }
    animElement(double angx, double angy)
    {
      anglex = angx; angley = angy;
    }
  };

  
  enum ShotOwner
  {
    SHOT_OWNER_NOBODY,
    SHOT_OWNER_ME,
    SHOT_OWNER_ENEMYSPERM
  };
  
  struct shotObj {
    ShotOwner owner;
    Point3D shotpos;
    Vector3D shotmove;
    Vector3D shotmovenorm;
    Vector3D originalMove;
    Vector3D shotacc;
    ParticleEngine* shot;
    bool shotdone;
    ParticleEngine::ParticleMode pmode;
    Uint32 texture;
    Colour col;
    bool collidable;
    double damage;

    shotObj() { }
    shotObj(Point3D pos, Vector3D vec, Vector3D acc, ShotOwner own, double damagePoint, int particlenum, double pvel, double pacc, Colour c, double colVariation, ParticleEngine::ParticleMode pm, Uint32 tex) 
    {
      damage = damagePoint;
      owner = own;
      collidable = true;
      shotpos = pos;
      shotmove = vec;
      shotmovenorm = shotmove;
      shotmovenorm.normalize();
      originalMove = shotmove;
      shotacc = acc;
      texture = tex;
      col = c;
      shot = new ParticleEngine(shotpos, pvel, pacc, particlenum, 0.005, col, colVariation, pm, texture);
      shotdone = false;
    }
    void tick() 
    {
      if (shot != NULL)
      {
        double percdone = shotmove.length2() / originalMove.length2() * 2;
	if (percdone > 1) percdone = 1;
	shot->setVelFactor(shot->getVelFactor() * percdone);
	shot->setPos(shotpos);
        shot->tick();
      }
      
      
      
      shotpos = shotpos + shotmove;

      Vector3D oldshotmove = shotmove;
      shotmove = shotmove + shotacc;
      if (shotmove[0] < 0 && oldshotmove[0] > 0 || shotmove[0] > 0 && oldshotmove[0] < 0) { shotacc[0] = 0; shotmove[0] = 0; shot->setLive(false); }
      if (shotmove[1] < 0 && oldshotmove[1] > 0 || shotmove[1] > 0 && oldshotmove[1] < 0) { shotacc[1] = 0; shotmove[1] = 0; shot->setLive(false); }
      if (shotmove[2] < 0 && oldshotmove[2] > 0 || shotmove[2] > 0 && oldshotmove[2] < 0) { shotacc[2] = 0; shotmove[2] = 0; shot->setLive(false); }

      
      
      
      if (shot != NULL)
      {
        if (shot->isDone())
        {
          delete shot;
	  shot = NULL;
	  shotdone = true;
	  collidable = false;
        }    
      }
    }
    
    void kill()
    {
      shotmove = Vector3D(0,0,0);
      shotacc = Vector3D(0,0,0);
      shot->setLive(false);
      collidable = false;
    }
    
    void render()
    {
          if (shot != NULL)
	    shot->render();
    }
    
  };

  enum PickupType
  {
    PICKUP_GLUCOSE
  };
  
  struct pickupObj {
    Point3D objpos;
    ParticleEngine* effect;
    bool done;
    ParticleEngine::ParticleMode pmode;
    Uint32 texture;
    Colour col;
    PickupType present;
    int life;
    bool collidable;

    pickupObj(PickupType pres, int lifeticks, Point3D pos,int particlenum, double pvel, double pacc, Colour c, double colVariation, ParticleEngine::ParticleMode pm, Uint32 tex) 
    {
      life = lifeticks;
      present = pres;
      objpos = pos;
      texture = tex;
      col = c;
      effect = new ParticleEngine(pos, pvel, pacc, particlenum, 0.005, col, colVariation, pm, texture);
      done = false;
      collidable = true;
    }
    void tick() 
    {
      if (life > 0)
      {
        life--;
        if (life == 0) kill();
      }
      
      if (effect != NULL)
      {
        effect->tick();
	if (effect->isDone())
        {
          delete effect;
	  effect = NULL;
	  done = true;
        }    
      }
    }
    
    void kill()
    {
      effect->setLive(false);
      collidable=false;
    }
    
    void render()
    {
      if (effect != NULL)
        effect->render();
    }
    
  };
  
  




struct Ova
{
  Point3D pos;
  Vector3D dir;
  
  Ova(Point3D p, Vector3D d)
  {
    pos = p;
    dir = d;
  }
  
  void tick()
  {
     pos = pos + dir;
  }
  
  void render()
  {
    glPushMatrix();
    glTranslated(pos[0], pos[1], pos[2]);
    
    glCallList(SPHERE);    
    
    glPopMatrix();

  }    
};


struct OvaSplitter {
  std::vector<Ova> rest;
  std::vector<Ova> move;
  
  int numtick;
  int maxtick;
  double rotx, roty;
  double rotxinc, rotyinc;
  OvaSplitter()
  {
    numtick = 0;
    maxtick = 75;
    rotx = 0;
    roty = 0;
    rotxinc = 0.5;
    rotyinc = -1;
  }
  
  bool stillMoving() { return move.size() > 0; }
  
  void add(Point3D init)
  {
    rest.push_back(Ova(init, Vector3D(0,0,0)));
  }
  
  
  void tick()
  {
    rotx = rotx + rotxinc;
    roty = roty + rotyinc;
    if (rotx > 360) rotx = rotx - 360;
    if (roty > 360) roty = roty - 360;
    if (rotx < 0) rotx = rotx + 360;
    if (roty < 0) roty = roty + 360;
    
    if (!move.empty())
    {
      numtick++;
      for (std::vector<Ova>::iterator it = move.begin(); it != move.end(); it++)
      {
         (*it).tick();    
      }      
    
      if (numtick > maxtick)
      {
        numtick = 0;
        for (std::vector<Ova>::iterator it = move.begin(); it != move.end(); it++)
        {
           rest.push_back((*it));
        }      
        move.clear();
      } 
    } 
  
  }
  
  void spawn()
  {
    std::vector< std::vector<Ova>::iterator > killList;
    for (std::vector<Ova>::iterator it = rest.begin(); it != rest.end(); it++)
    {
      Vector3D newdir(rand()%2 ==0 ? rand() : -rand(), rand()%2 ==0 ? rand() : -rand(), rand()%2 ==0 ? rand() : -rand());
      newdir.normalize();
      newdir = 0.01 * newdir;
      
      Ova o1((*it).pos, newdir);
      Ova o2((*it).pos, -newdir);
      move.push_back(o1);
      move.push_back(o2);
      killList.push_back(it);  
    }
    for (int i = 0; i < killList.size(); i++)
      rest.erase(killList[i]);
      
  }
  
  
  
  void quicksort(std::vector<Ova> &array, int top, int bottom)
  {
      // uses recursion-the process of calling itself
     int middle;
     if(top>bottom)
    {
          middle = partition(array, top, bottom);
          quicksort(array, top, middle);   // sort top partition
          quicksort(array, middle+1, bottom);    // sort bottom partition
     }
     return;
  }


  int partition(std::vector<Ova> &array, int top, int bottom)
  {
     double x = array[top].pos[2];
     int i = top - 1;
     int j = bottom + 1;
     Ova temp(Point3D(0,0,0), Vector3D(0,0,0));
     do
     {
           do     
           {
                  j--;
           }while (x >array[j].pos[2]);

          do  
         {
                 i++;
          } while (x <array[i].pos[2]);

          if (i < j)
         { 
                 temp = array[i];    // switch elements at positions i and j
                 array[i] = array[j];
                 array[j] = temp;
         }
     }while (i < j);    
     return j;           // returns middle index
  }
  
  
  
  
  
  
  
  
  
  
  
  void render()
  {
    
    
    //if (!move.empty()) quicksort(move, 0, move.size()-1);
    
    glPushMatrix();
      glRotated(rotx, 1, 0, 0);
      glRotated(roty, 0, 1, 0);
      
      glColor4d(1,1,1, 0.69);    
      if (!move.empty())
        glColor4d(1,1,1,((double)numtick/maxtick)*0.25 + 0.42);
      

      for (int i = 0; i < rest.size(); i++)
      {
        rest[i].render();
      }  
      for (int i = 0; i < move.size(); i++)
      {
        move[i].render();
      }  
    glPopMatrix();
  }
  
};




class Viewer
{

public:
  Viewer(int fullsc);

  bool intro();
  void outro();

  bool run();
  GLvoid glPrint( GLint x, GLint y, char *string, double scalef);
  void loadTexture(GLuint holder, std::string imageName, double numElem);

  std::map<std::string, int> soundMap;

  static std::vector<shotObj> shots;
  static std::vector<shotObj> shotSecondary;

  
private:

  
 


  int oldx, oldy;
  

  int filter;

  int music_on;

  double rotational_amt;

  Matrix4x4 camtrans;
  Matrix4x4 camtransInv;
  Point3D orientation;
  bool alive;
  double spermAlpha;
  
  double ammoAmt;
  double lifeAmt;
  double ammoRegenRate;
  double ammoAmtSlow;
  double lifeAmtSlow;

  GLuint  texture[20];

  bool wireframe;
  bool fog;

   
  World*env;
  World *env2;

  Uint32 curtime;


  GLuint  fontBase;

  Console gameConsole;

  int fullscreen;

  bool debug;

  int fpsCount;
  Uint32 fpsTimer;
  int fps;

  

  SceneNode* spermRoot;
  
  Point3D curpos;
  Vector3D curvel;
  
  
  void BuildFont();
  
  
  void DrawIntroScene();
  void DrawOutroScene();

  void makeMickey();
  void DrawGLScene();
  void handleKey(SDL_KeyboardEvent key);
  void handleMouseMotion(SDL_Event event);
  void runCommand(std::string cmd);
  void handleSmoothInput();
  inline void handleMouseButtons(SDL_MouseButtonEvent button);
  std::vector<std::string> explode(std::string str, std::string delimiters);

  bool loadKeys(std::string masterName, std::string filename);
  void interpolate(std::string masterName, std::string jointName);
  
  void vAxisRotMatrix(Vector3D vec, Matrix4x4 &mNewMat);
  void vCalcRotVec(double fNewX, double fNewY, double fOldX, double fOldY, double fDiameter, Vector3D &vec);
  
  void Viewer::displayIndicators();

 
  bool animFlag;
  int animCount;

  //map from MASTER level to map of frame numbers to map of joint names to animElements
  std::map<std::string, std::map<int, std::map<std::string, animElement> > > animMap;
  std::map<std::string, int> frameNumMap;
  void animTick(std::string masterName);
  
  Uint32 tailtick;
  

  
  Vector3D velocity;
  Vector3D rotVelocity;
  
  
  std::vector<pickupObj> powerups;
  
  int windowWidth;
  int windowHeight;
  
  bool alternateInput;
  
  bool showHelp;
  
  bool envCollide;
  
  std::vector<TextLine> helpText;
  std::vector<TextLine> creditText;
  
  
  double indicatorTextureOffset;
  
  int timeWaster;
  
  
  Matrix4x4 godViewRotate;
  double godViewX, godViewY, godViewZ;
  
  
  Uint32 deathTime;


  Cam cam;  
  
  
  
  
  bool showTexture;
  bool godView;
  bool godMode;
  bool musicEnabled;
  bool soundEnabled;
  bool frustumCull;
  
  
  bool intro_doneOpenScreen;
  double intro_openScreenAlpha;
  double outro_openScreenAlpha;

  Uint32 intro_timer;
  Uint32 intro_fadeTimer;

  double intro_scrollAmt;
  
  Uint32 outro_splitTimer;
  Uint32 outro_tickTimer;
  OvaSplitter outro_ova;

  GameSelectMenu gameMenu;
  
  
  std::vector<Enemy*> enemyList;

};




struct TextLine {
   
   enum PlaceMode {
     PLACE_EXACT,
     PLACE_ALIGN,
     PLACE_ALIGN_THEN_EXACT,
   };
   
   std::string text;
   int xpos, ypos;
   Colour col;
   Viewer *vu;
   PlaceMode pmode; 
   
   TextLine(std::string t, int x, int y, Colour c, PlaceMode p, Viewer *v)
   {
     text = t;
     xpos = x;
     if (p == PLACE_ALIGN || p == PLACE_ALIGN_THEN_EXACT)
       ypos = y*1*16;
     else
       ypos = y;
     col = c;
     vu = v;
     pmode = p;
   }
   TextLine(std::string t, int y, Colour c, PlaceMode p, Viewer *v)
   {
     
     text = t;
     xpos = (640 - 10*t.length()) / 2;
     if (p == PLACE_ALIGN || p == PLACE_ALIGN_THEN_EXACT)
       ypos = y*1*16;
     else
       ypos = y;
     col = c;
     vu = v;
     pmode = p;
   }

   void shiftX(int dx)
   {
     xpos = xpos + dx;
   }

   void shiftY(int dy)
   {
     if (pmode == PLACE_ALIGN)
       ypos = ypos + (dy*1*16);
     else
       ypos = ypos + dy;
   }

            
   void print()
   {
     glEnable(GL_TEXTURE_2D);
     glColor3d(col.R(),col.G(),col.B());
     vu->glPrint(xpos, 440-ypos, const_cast<char*>(text.c_str()), 1);
     glDisable(GL_TEXTURE_2D);
   }    
 
 };


 
 
 



#endif



