#ifndef ENV_XIX_H
#define ENV_XIX_H

#include "SDL_OGL.h" 
#include "image.hpp"
#include <stdlib.h>
#include <iostream>
#include "algebra.hpp"
#include <vector>
#include "a2.hpp"
#include <string>
#include "cam.hpp"
#include "material.hpp"

class Particle
{
  public:
    Particle(Point3D pos, Vector3D vel, Vector3D accel, double fade, Colour c);
    void sphereBillboardBegin(Point3D camPoint);
    void buildRot(float theta, float x, float y, float z, float m[16]);
    Particle();
    void render();
    void tick();
    bool alive() { return life > 0; }
    
  private:
  
    Vector3D vel;
    Vector3D accel;
    Point3D pos;
    double life;
    double size;
    double fade;
    Colour col;
    double sinTexAng;
    double cosTexAng;

};


class ParticleEngine
{
  public:
    
    enum ParticleMode {
     PARTICLE_FIRE,
     PARTICLE_EXPLODE
   };
    
    
    ParticleEngine(Point3D pos, double velFactor, double accFactor, int num, double fade, Colour c, double colVariation, ParticleMode pmode, GLuint texture);
    ~ParticleEngine();
    
    void render();
    void tick();
    void setLive(bool life) {alive = life;}
    bool isDone() { return done; }
    void setPos(Point3D newpos);
    void setCol(Colour c) {col = c;}
    void setVelFactor(double vel) { velFactor = vel; }
    double getVelFactor() { return velFactor; }
    void setAccFactor(double acc) { accFactor = acc; }
    double getAccFactor() { return accFactor; }
  private:
    Particle *particles;
    ParticleMode particleMode;
    int numParticle;
    Vector3D vel;
    Vector3D accel;
    Point3D pos;
    double fade;
    Colour col;
    double colVariation;
    GLuint texture;
    bool alive;
    bool done;
    double velFactor;
    double accFactor;
};


class WorldPortion
{

public:
  WorldPortion(int startX, int startY, int width, int height, double scale, Point3D **grid, Vector3D **normal, Vector3D **planeNormal, Point2D **textureCoord,  double *incAngle, Vector3D ***animatedNormals);
  ~WorldPortion();

  void render(Cam *c, int inc);
  bool intersect(Point3D pos, Vector3D dir, double rad, bool reverse, Vector3D &norm, double &dist);
  int intersect_triangle(Point3D orig, Vector3D dir, 
                   Point3D vert0, Point3D vert1, Point3D vert2, bool reverse, 
                   double *t, double *u, double *v, Vector3D &norm);
   
 

private:
    bool boundingVolumeIntersection(Point3D pos, Vector3D dir, double rad, double &dist) ;
   
    
    
    Point3D **faceList;
    Vector3D *normalList;
    int numFace;
    
    Point3D **grid;
    Vector3D **normal;
    Vector3D **planeNormal;
    Point2D **textureCoord;
    
    int inc;
    
    double *incAngle;
    Vector3D ***animatedNormals;

    double heartBeatTimer;
    bool doHeartBeat;
    double textureoffsetx;
    double textureoffsety;
    
    int envStartX;
    int envStartY;
    int envWidth;
    int envHeight;
    double m_scale;
    bool m_display;
    
    Point3D BBp1, BBp2;
    Point3D sphereCenter;
    double sphereRadius;
};



class World
{

public:
  World(std::string dispname, double scale, Matrix4x4 trans);
  ~World();

  void render(Cam *c, Uint32 outerTicks);
    
  bool intersect(Point3D pos, double rad);
  bool intersect(Point3D pos, Vector3D dir, double rad, bool reverse, Vector3D &norm, double &dist);

  static bool showNormal;
  static bool showBound;
  static bool animateNormal;
  static bool showFaceNormal;
  
  
private:
    Image displace;
    Point3D **grid;
    Vector3D **normal;
    Vector3D **planeNormal;
    Point2D **textureCoord;
    
    
    
    Vector3D getNormal(Point3D **grid, int i, int j, int width, int height);
    double m_scale;
    Point3D m_pos;
    int inc;
    
    double *incAngle;
    Vector3D ***animatedNormals;

    double heartBeatTimer;
    bool doHeartBeat;
    double textureoffsetx;
    double textureoffsety;
    Matrix4x4 m_trans;
    
    
    int envStartX;
    int envStartY;
    int envWidth;
    int envHeight;

    WorldPortion **ep;
    
  
};

















#endif
